import inspect
import multiprocessing as mp
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import partial
from numbers import Number
from typing import Callable, Dict, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from scipy import optimize as sop
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map

from dosma import defaults
from dosma.core.device import cpu_device, get_array_module, get_device
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences
from dosma.utils import env

if env.cupy_available():
    import cupy as cp

__all__ = [
    "CurveFitter",
    "PolyFitter",
    "MonoExponentialFit",
    "curve_fit",
    "polyfit",
    "monoexponential",
    "biexponential",
]


class _Fit(ABC):
    """Abstract class for fitting quantitative values."""

    @abstractmethod
    def fit(self) -> Tuple[MedicalVolume, MedicalVolume]:
        """Fit quantitative values per pixel across multiple volumes.

        Pixels with errors in fitting are set to np.nan.

        Returns:
            tuple[MedicalVolume, MedicalVolume]: Quantitative value volume and
                r-squared goodness of fit volume.
        """
        pass  # pragma: no cover


class _Fitter(ABC):
    # This is just a summary on variables used in this abstract class,
    # the proper values/initialization should be done in child class.
    nan_to_num: Optional[float]
    out_ufuncs: Optional[Union[Callable, Sequence[Callable]]]
    out_bounds: np.ndarray  # array-like
    r2_threshold: float
    y_bounds: Optional[Tuple[float]]

    def _format_out_ufuncs(self, _out_ufuncs, _func_nparams):
        if not isinstance(_out_ufuncs, Callable) and not all(
            isinstance(ufunc, Callable) or ufunc is None for ufunc in _out_ufuncs
        ):
            raise TypeError(
                f"`out_ufuncs` must be callable or sequence of callables. Got {_out_ufuncs}"
            )

        if isinstance(_out_ufuncs, Sequence):
            if len(_out_ufuncs) > _func_nparams:
                warnings.warn(
                    f"len(out_ufuncs)={len(_out_ufuncs)}, but only {_func_nparams} parameters. "
                    f"Extra ufuncs will be ignored."
                )

        return _out_ufuncs

    def _format_out_bounds(self, _out_bounds):
        out_bounds = np.asarray(_out_bounds)
        if out_bounds.shape[-1] != 2 or out_bounds.ndim > 2:
            raise ValueError("Invalid `out_bounds` - shape must be ([num_params,] 2)")
        if np.any(out_bounds[..., 0] > out_bounds[..., 1]):
            raise ValueError("Invalid `out_bounds` - lower bound must be <= upper bound")
        return out_bounds

    def _format_r2_threshold(self, _r2_threshold):
        if isinstance(_r2_threshold, str):
            if _r2_threshold != "preferences":
                raise ValueError(
                    f"Invalid value r2_threshold='{_r2_threshold}'. "
                    f"Expected `None`, a number between [0, 1], or 'preferences'."
                )
            _r2_threshold = preferences.fitting_r2_threshold
        return _r2_threshold

    def _process_mask(self, mask, y: MedicalVolume):
        """Process mask into appropriate shape."""
        arr_types = (np.ndarray, cp.ndarray) if env.cupy_available() else (np.ndarray,)
        if isinstance(mask, arr_types):
            mask = y._partial_clone(volume=mask, headers=None)
        elif not isinstance(mask, MedicalVolume):
            raise TypeError("`mask` must be a MedicalVolume or ndarray")

        mask = mask.reformat_as(y)
        if not mask.is_same_dimensions(y, defaults.AFFINE_DECIMAL_PRECISION):
            raise RuntimeError("`mask` and `y` dimension mismatch")

        return mask > 0

    def _process_params(self, x, r_squared):
        """
        Applies ``self.out_ufuncs`` and ``self.out_bounds``, in that order.
        Input is modified in place.

        Returns:
            ndarray: Values outside of ``self.out_bounds`` will be set to np.nan.
        """
        nan_to_num = self.nan_to_num
        out_ufuncs = self.out_ufuncs
        out_bounds = self.out_bounds
        r2_threshold = self.r2_threshold
        nparams = x.shape[-1]

        if isinstance(out_ufuncs, Callable):
            x = out_ufuncs(x)
        elif isinstance(out_ufuncs, Sequence):
            for i in range(min(nparams, len(out_ufuncs))):
                if out_ufuncs[i] is not None:
                    x[..., i] = out_ufuncs[i](x[..., i])

        if out_bounds is not None:
            if out_bounds.ndim == 2:
                extra_bounds = [(-np.inf, np.inf) for _ in range(nparams - out_bounds.shape[0])]
                if len(extra_bounds) > 0:
                    extra_bounds = np.stack(extra_bounds, axis=0)
                    out_bounds = np.concatenate([out_bounds, extra_bounds], axis=0)
                out_bounds = out_bounds.T
            lb, ub = out_bounds[0], out_bounds[1]
            x[(x < lb) | (x > ub)] = np.nan

        if r2_threshold is not None:
            x[(r_squared < r2_threshold)] = np.nan

        if nan_to_num is not None:
            x = np.nan_to_num(x, nan=nan_to_num, copy=False)

        return x

    def _fit(self, x, y):
        """Internal fitting function.

        Returns:
            tuple[array-like, array-like]: The fitted parameters and corresponding
                goodness of fit (:math:`r^2`).
        """
        raise NotImplementedError  # pragma: no cover

    def fit(self, x, y: Sequence[MedicalVolume], mask=None, copy_headers: bool = True, **kwargs):
        """
        Args:
            x (array-like): 1D array of independent variables corresponding to different ``y``.
            y (list[MedicalVolumes]): Dependent variable (in order) corresponding to values of
                independent variable ``x``. Data must be spatially aligned.
            mask (``MedicalVolume`` or ``ndarray``, optional): If specified, only entries where
                ``mask > 0`` will be fit.
            copy_headers (bool): If ``True``, headers will be deep copied. If ``False``,
                headers will not be copied. Returned values will not have headers.
            kwargs: Optional keyword arguments to be passed to :func:`self._fit`.

        Returns:
            Tuple[MedicalVolume, MedicalVolume]: Tuple of fitted parameters (``popt``)
            and goodness of fit (``r2``) values. Last axis of fitted parameters
            corresponds to different parameters in order of appearance in ``self.func``.
        """
        svs = []

        if (not isinstance(y, (list, tuple))) or (
            not all([isinstance(_y, MedicalVolume) for _y in y])
        ):
            raise TypeError("`y` must be sequence of MedicalVolumes.")

        x = np.asarray(x)
        if x.shape[-1] != len(y):
            raise ValueError(
                "Dimension mismatch: x.shape[-1]={:d}, but len(y)={:d}".format(x.shape[-1], len(y))
            )

        orientation = y[0].orientation
        y = [_y.reformat(orientation) for _y in y]

        if mask is not None:
            mask = self._process_mask(mask, y[0])
            mask = mask.volume.reshape(-1)

        original_shape = y[0].shape
        svs = np.concatenate([_y.volume.reshape((1, -1)) for _y in y], axis=0)
        flattened_shape = svs.shape

        # Select indices to fit.
        if mask is not None:
            svs = svs[:, mask]

        popt, r_squared = self._fit(x, svs, **kwargs)
        popt = self._process_params(popt, r_squared)

        if mask is not None:
            popt_full = np.empty(flattened_shape[-1:] + popt.shape[-1:])
            r2_full = np.empty(flattened_shape[-1])
            nan_val = np.nan if self.nan_to_num is None else self.nan_to_num
            popt_full.fill(nan_val)
            r2_full.fill(nan_val)
            popt_full[mask] = popt
            r2_full[mask] = r_squared
            popt = popt_full
            r_squared = r2_full
            del popt_full, r2_full  # variables not used later - drop reference to underlying array

        popt = popt.reshape(original_shape + popt.shape[-1:])
        r_squared = r_squared.reshape(original_shape)

        # For parameters, headers have to be deep copied and expanded to follow
        # broadcasting dimensions.
        if copy_headers:
            headers = y[0].headers()
            if headers is not None:
                headers = deepcopy(headers)
                if popt.ndim > y[0].volume.ndim:
                    axis = tuple(-i for i in range(1, popt.ndim - y[0].volume.ndim + 1))
                    headers = np.expand_dims(headers, axis=axis)
            popt_headers, r2_headers = headers, True
        else:
            popt_headers, r2_headers = None, None
        popt = y[0]._partial_clone(volume=popt, headers=popt_headers)
        rsquared_volume = y[0]._partial_clone(volume=r_squared, headers=r2_headers)

        return popt, rsquared_volume


class CurveFitter(_Fitter):
    """The class using non-linear least squares to fit a function to data.

    This class is a wrapper around the :func:`dosma.utils.fits.curve_fit` function
    that handles :class:`MedicalVolume` data and supports additional post-processing
    on fitted parameters.

    ``self.fit(x, y, mask=None)`` will fit independent variables ``x`` with observed
    MedicalVolumes ``y``. To only fit certain regions of volumes ``y``, specify ``mask``.

    Args:
        func (callable): The model function, f(x, ...). It must take the independent variable
            as the first argument and the parameters to fit as separate remaining arguments.
        p0 (Sequence, optional): Initial guess for the parameters (length N). If None, then
            the initial values will all be 1 (if the number of parameters for the
            function can be determined using introspection, otherwise a ValueError is raised).
        y_bounds (tuple, optional): Lower and upper bound on y values. Defaults to no bounds.
            Sequences with observations out of this range will not be processed.
        out_ufuncs (Callable, Sequence[Callable]): Function(s)
            to post-process parameter maps. Defaults to no post-processing.
            Each function must operate on ndarrays in an element-by-element fashion,
            should only take one argument, which will be the parameter map, and output
            the processed parameter map (an ndarray). If ``isinstance(out_ufuncs, Callable)``,
            operates on map of all parameters, where last dimension (``axis=-1``) corresponds to
            different parameters. If ``isinstance(out_ufuncs, Sequence)``, each ufunc corresponds
            to one parameter.
        out_bounds (array-like, optional): Lower and upper bounds (inclusive) on fitted parameters.
            Defaults to no bounds. Fitted parameters outside of this range will be set to
            ``np.nan``. Last dimension should have size 2. If ``out_bounds.ndim == 1``, bounds
            will be applied to all parameters.
        r2_threshold (float): Minimum :math:`r^2` goodness of fit value to accept fit per sample.
            Parameter values below this threshold will be set to np.nan.
            Defaults to ``preferences.fitting_r2_threshold``. To ignore, set to ``None``.
        nan_to_num (float, optional): If specified, all fitted parameters equal to ``np.nan``
            will be converted to this value. This uses ``numpy.nan_to_num``, which will also
            replace *posinf* and *neginf* values with very large positive and negative values,
            respectively. See ``numpy.nan_to_num`` for more details.
        num_workers (int, optional): Maximum number of workers to use for fitting.
        chunksize (int, optional): Size of chunks sent to worker processes when
            ``num_workers > 0``. When ``show_pbar=True``, this defaults to the standard
            value in :func:`tqdm.concurrent.process_map`.
        verbose (bool, optional): If `True`, show progress bar. Note this can increase runtime
            slightly when using multiple workers.
        kwargs: Keyword args for :func:`dosma.utils.fits.curve_fit`.

    Examples:
        Fitting :func:`monoexponential` (:math:`y = a * e^{b*x}`) to independent variables ``x``
        and dependent variables ``y`` with initial guesses ``a=1.0`` and ``b=-0.2``:

        >>> fitter = CurveFitter(monoexponential, p0=(1.0, -0.2))
        >>> popt, r2 = fitter.fit(x, y)
        >>> a_hat, b_hat = popt[..., 0], popt[..., 1]

        Post-process ``b`` by taking the inverse of its absolute value of ``b``
        (i.e. :math:`\\frac{1}{|b|}`). Set all values not in domain
        :math:`0 \\leq \\frac{1}{|b|} \\leq 100` or with a goodness of fit less than 0.9
        (:math:`r^2` < 0.9) to ``np.nan``:

        >>> ufunc = lambda x: 1 / np.abs(x)
        >>> out_bounds = ((-np.inf, np.inf), (0, 100))
        >>> fitter = CurveFitter(
        ...     monoexponential, p0=(1.0, -0.2), out_ufuncs=[None, ufunc], out_bounds=out_bounds)
        >>> popt, r2 = fitter.fit(x, y)
        >>> a_hat, inv_abs_b_hat = popt[..., 0], popt[..., 1]
    """

    def __init__(
        self,
        func: Callable,
        p0: Sequence[float] = None,
        y_bounds: Tuple[float] = None,
        out_ufuncs: Union[Callable, Sequence[Callable]] = None,
        out_bounds=None,
        r2_threshold: float = "preferences",
        nan_to_num: float = None,
        num_workers: int = 0,
        chunksize: int = None,
        verbose: bool = False,
        **kwargs,
    ):
        func_name = func.__name__ if hasattr(func, "__name__") else type(func).__name__
        sig = inspect.signature(func)
        func_args = list(sig.parameters)
        func_nparams = len(func_args) - 2 if "self" in func_args else len(func_args) - 1

        if out_ufuncs is not None:
            out_ufuncs = self._format_out_ufuncs(out_ufuncs, func_nparams)

        if out_bounds is not None:
            out_bounds = self._format_out_bounds(out_bounds)

        r2_threshold = self._format_r2_threshold(r2_threshold)

        self._func = func
        self._func_name = func_name
        self.p0 = self._format_p0(p0)
        self.y_bounds = y_bounds
        self.out_ufuncs = out_ufuncs
        self.out_bounds = out_bounds
        self.r2_threshold = r2_threshold
        self.nan_to_num = nan_to_num
        self.num_workers = num_workers
        self.chunksize = chunksize
        self.verbose = verbose
        self.kwargs = kwargs

    def _format_p0(
        self, p0, ref: MedicalVolume = None, flatten: bool = False, mask=None, depth: int = 0
    ):
        if p0 is None or isinstance(p0, Number):
            return p0
        elif isinstance(p0, MedicalVolume) and depth > 0:
            if ref is not None:
                p0 = p0.reformat_as(ref)
                assert p0.is_same_dimensions(ref, err=True)
            if flatten:
                p0 = p0.A.flatten()
                if mask is not None:
                    p0 = p0[mask]
            return p0
        elif isinstance(p0, np.ndarray) and depth > 0:
            if ref is not None and p0.shape != ref.shape:
                raise ValueError(f"Got p0.shape={p0.shape}, but y.shape={ref.shape}")
            if flatten:
                p0 = p0.flatten()
            if mask is not None:
                p0 = p0[mask]
            return p0

        if isinstance(p0, Mapping):
            return {k: self._format_p0(v, ref, flatten, mask, depth + 1) for k, v in p0.items()}
        elif isinstance(p0, Sequence):
            return tuple(self._format_p0(v, ref, flatten, mask, depth + 1) for v in p0)
        elif isinstance(p0, (np.ndarray, MedicalVolume)):
            # If a single numpy array is provided as the parameter input, the
            # last axis is considered to be the parameter axis. Note, this may
            # change in the future.
            return tuple(
                self._format_p0(p0[..., i], ref, flatten, mask, depth + 1)
                for i in range(p0.shape[-1])
            )

        raise ValueError(f"p0={p0} not supported")

    def fit(
        self, x, y: Sequence[MedicalVolume], mask=None, p0=np._NoValue, copy_headers: bool = True
    ):
        """Perform non-linear least squares fit.

        Args:
            x (array-like): 1D array of independent variables corresponding to different ``y``.
            y (list[MedicalVolumes]): Dependent variable (in order) corresponding to values of
                independent variable ``x``. Data must be spatially aligned.
            mask (``MedicalVolume`` or ``ndarray``, optional): If specified, only entries where
                ``mask > 0`` will be fit.
            p0 (Sequence, optional): Initial guess for the parameters.
                Defaults to ``self.p0``.
            copy_headers (bool, optional): If ``True``, headers will be deep copied. If ``False``,
                headers will not be copied. Returned values will not have headers.

        Returns:
            Tuple[MedicalVolume, MedicalVolume]: Tuple of fitted parameters (``popt``)
            and goodness of fit (``r2``) values. Last axis of fitted parameters
            corresponds to different parameters in order of appearance in ``self.func``.
        """
        if get_device(x) != cpu_device:
            raise RuntimeError("`x` must be on the CPU")
        if any(get_device(_y) != cpu_device for _y in y):
            raise RuntimeError("All elements in `y` must be on the CPU")

        if mask is not None:
            mask = self._process_mask(mask, y[0])

        if p0 is np._NoValue:
            p0 = self.p0
        p0 = self._format_p0(
            p0,
            ref=y[0],
            flatten=True,
            mask=mask.A.reshape(-1) if mask is not None else None,
        )

        return super().fit(x, y, mask=mask, p0=p0, copy_headers=copy_headers)

    def _fit(self, x, y, p0=np._NoValue):
        assert p0 is not np._NoValue

        return curve_fit(
            self._func,
            x,
            y,
            self.y_bounds,
            p0=p0,
            show_pbar=self.verbose,
            num_workers=self.num_workers,
            chunksize=self.chunksize,
            **self.kwargs,
        )

    def __str__(self) -> str:
        chunk_size = 1
        attrs = [
            "p0",
            "y_bounds",
            "out_bounds",
            "r2_threshold",
            "nan_to_num",
            "num_workers",
            "chunksize",
            "verbose",
        ]
        vals = [f"func={self._func_name}"]
        vals += [f"{k}={getattr(self, k)}" for k in attrs]
        vals += [f"{k}={v}" for k, v in self.kwargs.items()]
        val_str = "\n\t".join(
            [
                ", ".join(vals[i * chunk_size : (i + 1) * chunk_size]) + ","
                for i in range(int(np.ceil(len(vals) / chunk_size)))
            ]
        )
        return f"{self.__class__.__name__}(\n\t" + val_str + "\n)"


class PolyFitter(_Fitter):
    """The class using linear least squares to fit a polynomial to data.

    This class is a wrapper around the :func:`dosma.utils.fits.polyfit` function
    that handles :class:`MedicalVolume` data and supports additional post-processing
    on fitted parameters. Most attributes are shared with :class:`dosma.utils.fits.CurveFitter`.
    See that class for details.

    Unlike non-linear curve fitting, polynomial fitting can be done for multiple data sequences
    (i.e. multiple pixels/voxels) as a single least squares problem, which is the default in
    :mod:`numpy`. However, poor matrix conditioning can result in poor parameter estimates.
    As a result, this class also supports fitting each data sequence separately.
    If ``num_workers = None``, data sequences will be fit as a single least squares problem.
    If ``num_workers = 0``, data sequences will be fit separately and sequentially.
    If ``num_workers > 0``, multiprocessing will be used to fit each data sequence separately
    and in parallel.

    Args:
        deg (int): Degree of the fitting polynomial. See :func:`numpy.polyfit`.
        rcond (float, optional): Relative condition number of the fit. See :func:`numpy.polyfit`.
        y_bounds (tuple, optional): Same as :class:`CurveFitter`.
        out_ufuncs (Callable, Sequence[Callable]): Same as :class:`CurveFitter`.
        out_bounds (array-like, optional): Same as :class:`CurveFitter`.
        r2_threshold (float): Same as :class:`CurveFitter`.
        nan_to_num (float, optional): Same as :class:`CurveFitter`.
        num_workers (int, optional): Maximum number of workers to use for fitting.
            If ``None``, all data sequences should be solved as a single least squares problem.
            If ``0``, each data sequence will be fit separately from one another.
            Defaults to ``None``.
        chunksize (int, optional): Same as :class:`CurveFitter`.
        verbose (bool, optional): Same as :class:`CurveFitter`.

    Examples:
        Fitting polynomial of degree 1 (:math:`y = a*x + b`) to to independent variables ``x``
        and dependent variables ``y``.

        >>> fitter = PolyFitter(deg=1)
        >>> popt, r2 = fitter.fit(x, y)
        >>> a_hat, b_hat = popt[..., 0], popt[..., 1]

        Post-process ``b`` by taking the absolute value of ``b`` (i.e. :math:`|b|`).
        Set all values not in domain :math:`0 \\leq |b| \\leq 100` or with a goodness of fit
        less than 0.9 (:math:`r^2` < 0.9) to ``np.nan``:

        >>> ufunc = lambda x: np.abs(x)
        >>> out_bounds = ((-np.inf, np.inf), (0, 100))
        >>> fitter = CurveFitter(
        ...     monoexponential, p0=(1.0, -0.2), out_ufuncs=[None, ufunc], out_bounds=out_bounds)
        >>> popt, r2 = fitter.fit(x, y)
        >>> a_hat, abs_b_hat = popt[..., 0], popt[..., 1]
    """

    def __init__(
        self,
        deg: int,
        rcond: float = None,
        y_bounds: Tuple[float] = None,
        out_ufuncs: Union[Callable, Sequence[Callable]] = None,
        out_bounds=None,
        r2_threshold: float = "preferences",
        nan_to_num: float = None,
        num_workers: int = None,
        chunksize: int = None,
        verbose: bool = False,
    ):
        if out_ufuncs is not None:
            out_ufuncs = self._format_out_ufuncs(out_ufuncs, deg + 1)

        if out_bounds is not None:
            out_bounds = self._format_out_bounds(out_bounds)

        r2_threshold = self._format_r2_threshold(r2_threshold)

        self.deg = deg
        self.rcond = rcond

        self.y_bounds = y_bounds
        self.out_ufuncs = out_ufuncs
        self.out_bounds = out_bounds
        self.r2_threshold = r2_threshold
        self.nan_to_num = nan_to_num
        self.num_workers = num_workers
        self.chunksize = chunksize
        self.verbose = verbose

    def fit(self, x, y: Sequence[MedicalVolume], mask=None, copy_headers: bool = True):
        """Perform linear least squares fit.

        Args:
            x (array-like): Same as :meth:`CurveFitter.fit`.
            y (Sequence[MedicalVolume]): Same as :meth:`CurveFitter.fit`.
            mask (MedicalVolume or ndarray): Same as :meth:`CurveFitter.fit`.
            copy_headers (bool, optional): If ``True``, headers will be deep copied. If ``False``,
                headers will not be copied. Returned values will not have headers.

        Returns:
            Tuple[MedicalVolume, MedicalVolume]: Tuple of fitted parameters (``popt``)
                and goodness of fit (``r2``) values (same as `CurveFitter.fit`). The last
                axis of ``popt`` corresponds to different polynomial parameters in order
                ``y = popt[..., 0] * x ** self.deg + ... + popt[..., self.deg-1]``.
        """
        device = get_device(x)
        y_devices = [get_device(_y) for _y in y]
        if any(_y_device != device for _y_device in y_devices):
            raise RuntimeError(
                f"All elements in `y` must be on the same device as `x` ({device}). "
                f"Got {y_devices}."
            )

        return super().fit(x, y, mask=mask, copy_headers=copy_headers)

    def _fit(self, x, y):
        return polyfit(
            x,
            y,
            deg=self.deg,
            rcond=self.rcond,
            y_bounds=self.y_bounds,
            show_pbar=self.verbose,
            num_workers=self.num_workers,
            chunksize=self.chunksize,
        )

    def __str__(self) -> str:
        chunk_size = 1
        attrs = [
            "deg",
            "rcond",
            "y_bounds",
            "out_bounds",
            "r2_threshold",
            "nan_to_num",
            "num_workers",
            "chunksize",
            "verbose",
        ]
        vals = [f"{k}={getattr(self, k)}" for k in attrs]
        val_str = "\n\t".join(
            [
                ", ".join(vals[i * chunk_size : (i + 1) * chunk_size]) + ","
                for i in range(int(np.ceil(len(vals) / chunk_size)))
            ]
        )
        return f"{self.__class__.__name__}(\n\t" + val_str + "\n)"


class MonoExponentialFit(_Fit):
    """Fit quantitative values using mono-exponential fit of model :math:`y=a*exp(-x/tc)`.

    Args:
        x (:obj:`array-like`): 1D array of independent variables corresponding to different volumes.
        y (list[MedicalVolumes]): Volumes (in order) corresponding to independent variable in `x`.
        mask (:obj:`MedicalVolume`, optional): Mask of pixels to fit.
            If specified, pixels outside of mask region are ignored and set to ``np.nan``.
            Speeds fitting time as fewer fits are required.
        bounds (:obj:`tuple[float, float]`, optional): Upper and lower bound for quantitative
            values. Values outside those bounds will be set to ``np.nan``.
        tc0 (:obj:`float`, optional): Initial time constant guess. If ``"polyfit"``, this
            guess will be determined by first doing a polynomial fit on the log-linearized
            form of the monoexponential equation :math:`\\log y = \\log a - \\frac{t}{tc}`.
            Note for polyfit initialization, ``subvolumes`` should not have any nan or infinite
            values.
        decimal_precision (:obj:`int`, optional): Rounding precision after the decimal point.
        num_workers (int, optional): Maximum number of workers to use for fitting.
        chunksize (int, optional): Size of chunks sent to worker processes when
            ``num_workers > 0``. When ``show_pbar=True``, this defaults to the standard
            value in :func:`tqdm.concurrent.process_map`.
        verbose (bool, optional): If `True`, show progress bar. Note this can increase runtime
            slightly when using multiple workers.
    """

    def __init__(
        self,
        x: Sequence[float] = None,
        y: Sequence[MedicalVolume] = None,
        mask: MedicalVolume = None,
        bounds: Tuple[float] = (0, 100.0),
        tc0: Union[float, str] = 30.0,
        r2_threshold: float = "preferences",
        decimal_precision: int = 1,
        num_workers: int = 0,
        chunksize: int = 1000,
        verbose: bool = False,
    ):

        self.x = x
        if y is not None:
            warnings.warn(
                f"Setting `y` in the constructor can result in significant memory overhead. "
                f"Specify `y` in `{type(self).__name__}.fit(y=...)` instead."
            )
            self._check_y(x, y)
        self.y = y

        if mask is not None:
            warnings.warn(
                f"Setting `mask` in the constructor can result in significant memory overhead. "
                f"Specify `mask` in `{type(self).__name__}.fit(mask=...)` instead."
            )
        self.mask = mask

        if not (isinstance(tc0, Number) or (isinstance(tc0, str) and tc0 == "polyfit")):
            raise ValueError("`tc0` must either be a float or the string 'polyfit'.")

        self.verbose = verbose
        self.num_workers = num_workers

        if len(bounds) != 2:
            raise ValueError("`bounds` should provide lower/upper bound in format (lb, ub)")
        self.bounds = bounds

        self.chunksize = chunksize
        self.r2_threshold = r2_threshold
        self.tc0 = tc0
        self.decimal_precision = decimal_precision
        self._eps = 1e-10  # epsilon for polyfit - do not change this

    def fit(self, x=None, y: Sequence[MedicalVolume] = None, mask=None):
        """Perform monoexponential fitting.

        Returns:
            Tuple[MedicalVolume, MedicalVolume]:

                time_constant_volume (MedicalVolume): The per-voxel tc fit.
                rsquared_volume (MedicalVolume): The per-voxel r2 goodness of fit.
        """
        x = self.x if x is None else x
        y = self.y if y is None else y
        mask = self.mask if mask is None else mask

        self._check_y(x, y)
        orientation = y[0].orientation
        y = [sv.reformat(orientation) for sv in y]

        if isinstance(mask, np.ndarray):
            mask = MedicalVolume(mask, affine=y[0].affine)
            if not isinstance(mask, MedicalVolume):
                raise TypeError("`mask` must be a MedicalVolume")
        mask = mask.reformat(orientation) if mask else None

        if self.tc0 == "polyfit":
            polyfitter = PolyFitter(
                1,
                r2_threshold=0,
                num_workers=None,
                nan_to_num=0.0,
                chunksize=self.chunksize,
                verbose=self.verbose,
            )
            vols = [
                sv.astype(np.float32) if np.issubdtype(sv.dtype, np.integer) else sv for sv in y
            ]
            vols = [sv + self._eps * (sv == 0) for sv in vols]
            assert all(np.all(v != 0) for v in vols)
            vols = [np.log(v) for v in vols]
            params, _ = polyfitter.fit(x, vols, mask=mask, copy_headers=False)
            p0 = {"a": np.exp(params[..., 1]), "b": params[..., 0]}
            del vols  # begin garbage collection for large arrays sooner
        else:
            p0 = {"a": 1.0, "b": -1 / self.tc0}

        curve_fitter = CurveFitter(
            monoexponential,
            y_bounds=None,
            out_ufuncs=(None, lambda _x: 1 / np.abs(_x)),
            out_bounds=((-np.inf, np.inf), self.bounds),
            r2_threshold=self.r2_threshold,
            num_workers=self.num_workers,
            chunksize=self.chunksize,
            verbose=self.verbose,
            nan_to_num=0.0,
        )
        popt, r_squared = curve_fitter.fit(x, y, mask=mask, p0=p0)
        tc_map = popt[..., 1]

        if self.decimal_precision is not None:
            tc_map = np.around(tc_map, self.decimal_precision)

        return tc_map, r_squared

    def _check_y(self, x, y):
        if (not isinstance(y, Sequence)) or (not all([isinstance(sv, MedicalVolume) for sv in y])):
            raise TypeError("`y` must be list of MedicalVolumes.")

        if any(x.device != cpu_device for x in y):
            raise RuntimeError("All MedicalVolumes must be on the CPU")

        if len(x) != len(y):
            raise ValueError("`len(x)`={:d}, but `len(y)`={:d}".format(len(x), len(y)))


__EPSILON__ = 1e-8


def curve_fit(
    func,
    x,
    y,
    y_bounds=None,
    p0=None,
    maxfev=100,
    ftol=1e-5,
    eps=1e-8,
    show_pbar=False,
    num_workers=0,
    chunksize: int = None,
    **kwargs,
):
    """Use non-linear least squares to fit a function ``func`` to data.

    Uses :func:`scipy.optimize.curve_fit` backbone.

    Args:
        func (callable): The model function, f(x, ...). It must take the independent variable
            as the first argument and the parameters to fit as separate remaining arguments.
        x (ndarray): The independent variable(s) where the data is measured.
            Should usually be an M-length sequence or an (k,M)-shaped array for functions
            with k predictors, but can actually be any object.
        y (ndarray): The dependent data, a length M array - nominally func(xdata, ...) - or
            an (M,N)-shaped array for N different sequences.
        y_bounds (tuple, optional): Lower and upper bound on y values. Defaults to no bounds.
            Sequences with observations out of this range will not be processed.
        p0 (Number | Sequence[Number] | ndarray | Dict, optional): Initial guess for the parameters.
            If sequence (e.g. list, tuple, 1d ndarray), it should have length P, which is the
            number of parameters. If this is a 2D numpy array, it should have a shape ``(N, P)``.
            If ``None``, then the initial values will all be 1.
        maxfev (int, optional): Maximum number of function evaluations before the termination.
            If `bounds` argument for `scipy.optimize.curve_fit` is specified, this corresponds
            to the `max_nfev` in the least squares algorithm
        ftol (float): Tolerance for termination by the change of the cost function.
            See `scipy.optimize.least_squares` for more details.
        eps (float, optional): Epsilon for computing r-squared.
        show_pbar (bool, optional): If `True`, show progress bar. Note this can increase runtime
            slightly when using multiple workers.
        num_workers (int, optional): Maximum number of workers to use for fitting.
        chunksize (int, optional): Size of chunks sent to worker processes when
            ``num_workers > 0``. When ``show_pbar=True``, this defaults to the standard
            value in :func:`tqdm.concurrent.process_map`.
        kwargs: Keyword args for `scipy.optimize.curve_fit`.

    Returns:
        Tuple[ndarray, ndarray]:

            popts (ndarray): A NxP matrix of fitted values. The last dimension (``axis=-1``)
            corresponds to the different parameters (in order).

            rsquared (ndarray): A (N,) length matrix of r-squared goodness-of-fit values.
    """
    if (get_device(x) != cpu_device) or (get_device(y) != cpu_device):
        raise RuntimeError("`x` and `y` must be on CPU")

    x = np.asarray(x)
    y = np.asarray(y)
    if y.ndim == 1:
        y = y.reshape(y.shape + (1,))
    N = y.shape[-1]

    func_args = list(inspect.signature(func).parameters)
    nparams = len(func_args) - 2 if "self" in func_args else len(func_args) - 1
    param_args = func_args[2:] if "self" in func_args else func_args[1:]

    if p0 is None:
        p0_scalars, p0_seq = None, None
    else:
        p0_scalars, p0_seq = _format_p0(p0, param_args, N)

    if "bounds" not in kwargs:
        kwargs["maxfev"] = maxfev
    elif "max_nfev" not in kwargs:
        kwargs["max_nfev"] = maxfev

    num_workers = min(num_workers, N)
    fitter = partial(
        _curve_fit,
        x=x,
        func=func,
        y_bounds=y_bounds,
        p0=p0_scalars,
        ftol=ftol,
        eps=eps,
        nparams=nparams,
        **kwargs,
    )

    oob = y_bounds is not None and ((y < y_bounds[0]).any() or (y > y_bounds[1]).any())
    if oob:
        warnings.warn("Out of bounds values found. Failure in fit will result in np.nan")

    y_T = y.T
    if p0_seq:
        y_T = [{"y": y_T[i], "p0": {k: v[i] for k, v in p0_seq.items()}} for i in range(N)]

    popts = []
    r_squared = []
    if not num_workers:
        for i in tqdm(range(N), disable=not show_pbar):
            popt_, r2_ = fitter(y_T[i])
            popts.append(popt_)
            r_squared.append(r2_)
    else:
        if show_pbar:
            tqdm_kwargs = {"chunksize": chunksize}
            tqdm_kwargs = {k: v for k, v in tqdm_kwargs.items() if v is not None}
            data = process_map(fitter, y_T, max_workers=num_workers, **tqdm_kwargs)
        else:
            with mp.Pool(num_workers) as p:
                data = p.map(fitter, y_T, chunksize=chunksize)
        popts, r_squared = [x[0] for x in data], [x[1] for x in data]

    return np.stack(popts, axis=0), np.asarray(r_squared)


def polyfit(
    x,
    y,
    deg: int,
    rcond=None,
    full=False,
    w=None,
    cov=False,
    eps=1e-8,
    y_bounds=None,
    show_pbar=False,
    num_workers=None,
    chunksize: int = None,
):
    """Use linear least squares to fit a polynomial of degree ``deg`` to data.

    This function is a wrapper around the :func:`numpy.polyfit` and :func:`cupy.polyfit`
    functions.

    In addition to standard polyfit functionality, this function also supports
    multiprocessing of the data using multiple workers. When multiprocessing is enabled,
    each data sequence is fit using a separate worker.

    In most cases, multiprocessing is not needed for linear least squares as most
    computations can be done quickly via matrix decomposition. However, there are cases
    where ``y`` values can cause low condition numbers and/or invalid outputs. In these cases,
    solving for each data sequence separately can be beneficial.

    Solving each data sequence separately without multiprocessing can also be done by setting
    ``num_workers=0``. This can be useful when certain data sequences are ill conditioned.

    Args:
        x (ndarray): The independent variable(s) where the data is measured.
            Should usually be an M-length sequence or an (k,M)-shaped array for functions
            with k predictors, but can actually be any object.
        y (ndarray): The dependent data, a length M array - nominally func(xdata, ...) - or
            an (M,N)-shaped array for N different sequences.
        deg (int): Degree of the fitting polynomial. Same as :func:`numpy.polyfit`.
        rcond (float, optional): Same as :func:`numpy.polyfit`.
        full (bool, optional): Same as :func:`numpy.polyfit`.
        w (array-like, optional): Same as :func:`numpy.polyfit`.
        cov (bool, optional): Same as :func:`numpy.polyfit`.
        eps (float, optional): Epsilon for computing r-squared.
        y_bounds (tuple, optional): Same as :func:`curve_fit`.
        show_pbar (bool, optional): Same as :func:`curve_fit`.
        num_workers (int, optional): Maximum number of workers to use for fitting.
            If ``None``, all data sequences should be solved as a single least squares problem.
            If ``0``, each data sequence will be fit separately from one another.
            Defaults to ``None``.
        chunksize (int, optional): Same as :func:`curve_fit`.
            Only used when ``num_workers`` is not ``None``.
    """

    def _compute_r2_matrix(_x, _y, _popts):
        """
        Here, ``M`` refers to # sample points, ``K`` refers to # sequences
        This function needs to be run under the correct context
        Args:
            _x: array_like, shape (M,)
            _y: array_like, shape (M, K)
            _popts (array-like): Shape (deg+1, K)
        """
        xp = get_array_module(_y)

        _x = _x.flatten()
        _xs = xp.stack([_x ** i for i in range(len(_popts) - 1, -1, -1)], axis=-1)
        yhat = _xs @ _popts  # (M, K)

        residuals = yhat - _y
        ss_res = xp.sum(residuals ** 2, axis=0)
        ss_tot = xp.sum((_y - xp.mean(_y, axis=0, keepdims=True)) ** 2, axis=0)
        return 1 - (ss_res / (ss_tot + eps))

    x_device, y_device = get_device(x), get_device(y)
    if x_device != y_device:
        raise ValueError(f"`x` ({x_device}) and `y` ({x_device}) must be on the same device")

    scatter_data = num_workers is not None
    if (cov or full) and scatter_data:
        raise ValueError("`cov` or `full` cannot be used with multiprocessing")

    xp = get_array_module(x)

    x = xp.asarray(x)
    y = xp.asarray(y)
    if y.ndim == 1:
        y = y.reshape(y.shape + (1,))
    N = y.shape[-1]

    num_workers = min(num_workers, N) if num_workers is not None else None

    oob = y_bounds is not None and ((y < y_bounds[0]).any() or (y > y_bounds[1]).any())
    if oob:
        warnings.warn("Out of bounds values found. Failure in fit will result in np.nan")

    fitter = partial(
        _polyfit, x=x, deg=deg, y_bounds=y_bounds, rcond=rcond, w=w, eps=eps, xp=xp.__name__
    )

    residuals, rank, singular_values, rcond, V = None, None, None, None, None
    with x_device:
        if not scatter_data:
            # Fit all sequences as one least squares problem.
            out = xp.polyfit(x, y, deg, rcond=rcond, full=full, w=w, cov=cov)
            if full:
                popts, residuals, rank, singular_values, rcond = out
            elif cov:
                popts, V = out
            else:
                popts = out
            r_squared = _compute_r2_matrix(x, y, popts)
            popts = popts.T
        elif num_workers == 0:
            # Fit each sequence separately
            popts = []
            r_squared = []
            for i in tqdm(range(N), disable=not show_pbar):
                popt_, r2_ = fitter(y[:, i])
                popts.append(popt_)
                r_squared.append(r2_)
            popts = xp.stack(popts, axis=0)
        else:
            # Multiprocessing - fits each sequence separately
            if show_pbar:
                tqdm_kwargs = {"chunksize": chunksize}
                tqdm_kwargs = {k: v for k, v in tqdm_kwargs.items() if v is not None}
                data = process_map(fitter, y.T, max_workers=num_workers, **tqdm_kwargs)
            else:
                with mp.Pool(num_workers) as p:
                    data = p.map(fitter, y.T, chunksize=chunksize)
            popts, r_squared = [x[0] for x in data], [x[1] for x in data]
            popts = xp.stack(popts, axis=0)

        r_squared = xp.asarray(r_squared)

    if full:
        return popts, r_squared, residuals, rank, singular_values, rcond
    elif cov:
        return popts, r_squared, V
    else:
        return popts, r_squared


def monoexponential(x, a, b):
    """Function: :math:`f(x) = a * e^{b*x}`."""
    return a * np.exp(b * x)


def biexponential(x, a1, b1, a2, b2):
    """Function: :math:`f(x) = a1*e^{b1*x} + a2*e^{b2*x}`."""
    return a1 * np.exp(b1 * x) + a2 * np.exp(b2 * x)


def _curve_fit(
    y_or_dict, x, func, y_bounds=None, p0=None, ftol=1e-5, eps=1e-8, nparams=None, **kwargs
):
    def _fit_internal(_x, _y):
        popt, _ = sop.curve_fit(func, _x, _y, p0=p0, ftol=ftol, **kwargs)

        residuals = _y - func(_x, *popt)
        ss_res = np.sum(residuals ** 2)
        ss_tot = np.sum((_y - np.mean(_y)) ** 2)
        r_squared = 1 - (ss_res / (ss_tot + eps))

        return popt, r_squared

    def _parse_p0(_dict):
        if "p0" not in _dict or _dict["p0"] in [None, {}]:
            return p0
        _p0_dict = _dict["p0"]
        if p0 is None:
            return tuple(_p0_dict.values())
        elif isinstance(p0, Dict):
            # In Python>=3.6 dictionary key/value pairs are ordered by default.
            # The order of the keys should reflect the order of the parameters
            # in the target function.
            p0_copy = p0.copy()
            p0_copy.update(_p0_dict)
            return tuple(p0_copy.values())
        else:
            assert False  # p0 must be None or a mapping

    if nparams is None:
        func_args = inspect.getargspec(func).args
        nparams = len(func_args) - 2 if "self" in func_args else len(func_args) - 1

    if isinstance(y_or_dict, dict):
        y = y_or_dict["y"]
        p0 = _parse_p0(y_or_dict)
    else:
        y = y_or_dict

    oob = y_bounds is not None and ((y < y_bounds[0]).any() or (y > y_bounds[1]).any())
    if oob or (y == 0).all():
        return (np.nan,) * nparams, 0

    try:
        popt_, r2_ = _fit_internal(x, y)
    except RuntimeError:
        popt_, r2_ = (np.nan,) * nparams, 0
    return popt_, r2_


def _polyfit(y, x, deg, y_bounds=None, xp=None, rcond=None, w=None, eps=1e-8):
    def _fit_internal(_x, _y):
        popt = xp.polyfit(_x, _y, deg, rcond=rcond, w=w)

        residuals = _y - xp.polyval(popt, _x)
        ss_res = xp.sum(residuals ** 2)
        ss_tot = xp.sum((_y - xp.mean(_y)) ** 2)
        r_squared = 1 - (ss_res / (ss_tot + eps))

        return popt, r_squared

    if xp in ["np", "numpy"]:
        xp = np
    elif xp in ["cp", "cupy"]:
        xp = cp
    else:
        raise ValueError("Cannot determine array module")

    nparams = deg + 1
    oob = y_bounds is not None and ((y < y_bounds[0]).any() or (y > y_bounds[1]).any())
    if oob or (y == 0).all():
        return (xp.nan,) * nparams, 0

    try:
        popt_, r2_ = _fit_internal(x, y)
    except RuntimeError:
        popt_, r2_ = (xp.nan,) * nparams, 0
    return popt_, r2_


def _format_p0(p0, param_args, N):
    """Formats the p0 values for :func:`curve_fit`.

    Args:
        p0: The initialization values for the parameters.
        param_args (Sequence[str]): The parameter names.
        N (int): The number of sequences.

    Returns:
        p0_scalars (Dict[str, Optional[Number]]): The dictionary of
            all paramters and their scalar initializations. If parameter
            does not have a scalar value, the value for that key will be
            ``None``.
        p0_seq (Dict[str, Union[Sequence, ndarray]]): The dictionary of
            partial parameters that have a sequence-like structure.
    """
    nparams = len(param_args)

    if isinstance(p0, Number):
        p0 = (p0,) * nparams
    elif isinstance(p0, np.ndarray) and p0.ndim > 1:
        p0 = tuple(p0[..., i] for i in range(p0.shape[-1]))

    if isinstance(p0, (np.ndarray, Sequence)):
        if len(p0) != nparams:
            raise ValueError(f"`p0` has length {len(p0)} but function has {nparams} parameters")
        p0 = {param_args[i]: p0[i] for i in range(nparams)}
    elif isinstance(p0, Mapping):
        extra_keys = set(p0) - set(param_args)
        if len(extra_keys) > 0:
            raise ValueError(
                f"`p0` has unknown keys: {extra_keys}. "
                f"Function signature has parameters {param_args}."
            )
        p0_default = {p: 1.0 for p in param_args}
        # Update p0_default to keep the parameter keys in order.
        p0_default.update(p0)
        p0 = p0_default

    if p0 is not None:
        p0.update({k: 1.0 for k, v in p0.items() if v is None})

        p0_scalars = {k: v if not isinstance(v, np.ndarray) else None for k, v in p0.items()}
        p0_seq = {k: v for k, v in p0.items() if isinstance(v, np.ndarray)}
        for k, v in p0_seq.items():
            if len(v) != N:
                raise ValueError(f"Got {len(v)} values for param '{k}'. Expected {N}")
        if not p0_scalars:
            p0_scalars = None
        if not p0_seq:
            p0_scalars = tuple(p0_scalars.values())
            p0_seq = None
    else:
        p0_scalars, p0_seq = None, None

    return p0_scalars, p0_seq


def __fit_mono_exp__(x, y, p0=None):  # pragma: no cover
    def func(t, a, b):
        exp = np.exp(b * t)
        return a * exp

    warnings.warn(
        "__fit_mono_exp__ is deprecated since v0.0.12 and will no longer be "
        "supported in v0.0.13. Use `curve_fit` instead.",
        DeprecationWarning,
    )

    x = np.asarray(x)
    y = np.asarray(y)

    popt, _ = sop.curve_fit(func, x, y, p0=p0, maxfev=100, ftol=1e-5)

    residuals = y - func(x, popt[0], popt[1])
    ss_res = np.sum(residuals ** 2)
    ss_tot = np.sum((y - np.mean(y)) ** 2)

    r_squared = 1 - (ss_res / (ss_tot + __EPSILON__))

    return popt, r_squared


def __fit_monoexp_tc__(x, ys, tc0, show_pbar=False):  # pragma: no cover
    warnings.warn(
        "__fit_monoexp_tc__ is deprecated since v0.12 and will no longer be "
        "supported in v0.13. Use `curve_fit` instead.",
        DeprecationWarning,
    )

    p0 = (1.0, -1 / tc0)
    time_constants = np.zeros([1, ys.shape[-1]])
    r_squared = np.zeros([1, ys.shape[-1]])

    warned_negative = False
    for i in tqdm(range(ys.shape[1]), disable=not show_pbar):
        y = ys[..., i]
        if (y < 0).any() and not warned_negative:
            warned_negative = True
            warnings.warn(
                "Negative values found. Failure in monoexponential fit will result in np.nan"
            )

        # Skip any negative values or all values that are 0s
        if (y < 0).any() or (y == 0).all():
            continue

        try:
            params, r2 = __fit_mono_exp__(x, y, p0=p0)
            tc = 1 / abs(params[-1])
        except RuntimeError:
            tc, r2 = (np.nan, 0.0)

        time_constants[..., i] = tc
        r_squared[..., i] = r2

    return time_constants, r_squared



"""Ultra-short Echo Time Cones (UTE-Cones)."""
import logging
import os
from typing import Sequence

import numpy as np

from dosma import file_constants as fc
from dosma.core import quant_vals as qv
from dosma.core.fitting import MonoExponentialFit
from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.nifti_io import NiftiReader
from dosma.core.med_volume import MedicalVolume
from dosma.core.registration import apply_warp, register
from dosma.scan_sequences.scans import NonTargetSequence
from dosma.tissues.tissue import Tissue
from dosma.utils.cmd_line_utils import ActionWrapper

__all__ = ["Cones"]

__EXPECTED_NUM_ECHO_TIMES__ = 4

__INITIAL_T2_STAR_VAL__ = 30.0

__T2_STAR_LOWER_BOUND__ = 0
__T2_STAR_UPPER_BOUND__ = np.inf
__T2_STAR_DECIMAL_PRECISION__ = 3

_logger = logging.getLogger(__name__)


class Cones(NonTargetSequence):
    """UTE-Cones MRI sequence.

    Ultra-short echo time cones (UTE-Cones) is a :math:`T_2^*`-weighted sequence.
    In practice, many of these scans are low resolution and are ofter interregistered
    with higher-resolution scans. This can be done with :meth:`Cones.interregister`.

    References:
        Qian Y, Williams AA, Chu CR, Boada FE. Multicomponent T2* mapping of
        knee cartilage: technical feasibility ex vivo.
        Magnetic resonance in medicine 2010;64(5):1426-1431."
    """

    NAME = "cones"

    def __init__(self, volumes, echo_times: Sequence[float] = None):
        super().__init__(volumes)

        if echo_times is None:
            try:
                if all(x.headers() is not None for x in self.volumes):
                    echo_times = [x.get_metadata("EchoTime", float) for x in self.volumes]
            except (KeyError, AttributeError, RuntimeError) as e:
                raise ValueError(
                    f"Could not extract echo times from header. "
                    f"Please specify `echo_times` argument - {e}"
                )

        self.echo_times = echo_times

    def interregister(self, target_path: str, target_mask_path: str = None):
        volumes = self.volumes
        echo_times = self.echo_times
        idxs = np.argsort(echo_times)

        echo_times = [echo_times[i] for i in idxs]
        volumes = [volumes[i] for i in idxs]
        nr = NiftiReader()
        out_path = os.path.join(self.temp_path, "interregistered")
        os.makedirs(out_path, exist_ok=True)

        # TODO: Make these into parameters
        num_threads = 2
        num_workers = 0
        verbose = True

        if verbose:  # pragma: no cover
            _logger.info("")
            _logger.info("==" * 40)
            _logger.info("Interregistering...")
            _logger.info("Target: {}".format(target_path))
            if target_mask_path is not None:
                _logger.info("Mask: {}".format(target_mask_path))
            _logger.info("==" * 40)

        # Target mask path has to be dilated.
        if target_mask_path:
            target_mask_path = self.__dilate_mask__(target_mask_path, out_path)
            parameter_files = [
                fc.ELASTIX_RIGID_INTERREGISTER_PARAMS_FILE,
                fc.ELASTIX_AFFINE_INTERREGISTER_PARAMS_FILE,
            ]
            use_mask = [False, True]
        else:
            parameter_files = [fc.ELASTIX_RIGID_PARAMS_FILE, fc.ELASTIX_AFFINE_PARAMS_FILE]
            use_mask = None

        # Last echo should be the base.
        base, moving = volumes[-1], volumes[:-1]

        out_reg, _ = register(
            target_path,
            base,
            parameters=parameter_files,
            output_path=out_path,
            sequential=True,
            collate=True,
            num_workers=num_workers,
            num_threads=num_threads,
            return_volumes=False,
            target_mask=target_mask_path,
            use_mask=use_mask,
            rtype=tuple,
            show_pbar=verbose,
        )
        out_reg = out_reg[0]

        reg_vols = []
        for mvg in moving:
            reg_vols.append(apply_warp(mvg, out_reg.transform))
        reg_vols.append(nr.load(out_reg.warped_file))  # base volume is last

        # Undo sorting by echo time.
        reverse_idxs = {v: i for i, v in enumerate(idxs)}
        reg_vols = [reg_vols[reverse_idxs[k]] for k in sorted(reverse_idxs.keys())]

        self.volumes = reg_vols

    def generate_t2_star_map(self, tissue: Tissue, mask_path: str = None, num_workers: int = 0):
        """
        Generate 3D :math:`T_2^* map and r-squared fit map using mono-exponential fit
        across subvolumes acquired at different echo times.

        :math:`T_2^* map is also added to the tissue.

        Args:
            tissue (Tissue): Tissue to generate quantitative value for.
            mask_path (:obj:`str`, optional): File path to mask of ROI to analyze.
                If specified, only voxels specified by mask will be fit.
                This can considerably speed up computation.
            num_workers (int, optional): Number of subprocesses to use for fitting.
                If `0`, will execute on the main thread.

        Returns:
            qv.T2Star: :math:`T_2^* fit for tissue.

        Raises:
            ValueError: If ``mask_path`` corresponds to non-binary volume.
        """
        # only calculate for focused region if a mask is available, this speeds up computation
        mask = tissue.get_mask()
        if mask_path is not None:
            mask = (
                fio_utils.generic_load(mask_path, expected_num_volumes=1)
                if isinstance(mask_path, (str, os.PathLike))
                else mask_path
            )

        spin_lock_times = self.echo_times
        subvolumes_list = self.volumes

        mef = MonoExponentialFit(
            bounds=(__T2_STAR_LOWER_BOUND__, __T2_STAR_UPPER_BOUND__),
            tc0="polyfit",
            decimal_precision=__T2_STAR_DECIMAL_PRECISION__,
            num_workers=num_workers,
            verbose=True,
        )

        t2star_map, r2 = mef.fit(spin_lock_times, subvolumes_list, mask=mask)

        quant_val_map = qv.T2Star(t2star_map)
        quant_val_map.add_additional_volume("r2", r2)

        tissue.add_quantitative_value(quant_val_map)

        return quant_val_map

    def _save(self, metadata, save_dir, fname_fmt=None, **kwargs):
        default_fmt = {MedicalVolume: "echo-{}"}
        default_fmt.update(fname_fmt if fname_fmt else {})
        return super()._save(metadata, save_dir, fname_fmt=default_fmt, **kwargs)

    @classmethod
    def from_dict(cls, data, force: bool = False) -> "Cones":
        interregistered_dirpath = None
        if "subvolumes" in data:
            interregistered_dirpath = os.path.dirname(data.pop("subvolumes")[0])
        scan: Cones = super().from_dict(data, force=force)
        if interregistered_dirpath is not None:
            subvolumes = scan.__load_interregistered_files__(interregistered_dirpath)
            cls.volumes = [subvolumes[k] for k in sorted(subvolumes.keys())]

        return scan

    @classmethod
    def cmd_line_actions(cls):
        """
        Provide command line information (such as name, help strings, etc)
        as list of dictionary.
        """

        interregister_action = ActionWrapper(
            name=cls.interregister.__name__,
            help="register to another scan",
            param_help={
                "target_path": "path to target image in nifti format (.nii.gz)",
                "target_mask_path": "path to target mask in nifti format (.nii.gz)",
            },
            alternative_param_names={
                "target_path": ["tp", "target"],
                "target_mask_path": ["tm", "target_mask"],
            },
        )
        generate_t2star_map_action = ActionWrapper(
            name=cls.generate_t2_star_map.__name__,
            help="generate T2-star map",
            param_help={
                "mask_path": "Mask used for fitting select voxels - " "in nifti format (.nii.gz)"
            },
            aliases=["t2_star"],
        )

        return [
            (cls.interregister, interregister_action),
            (cls.generate_t2_star_map, generate_t2star_map_action),
        ]



import logging
import os
from copy import deepcopy
from typing import List, Sequence

from nipype.interfaces.elastix import Registration

from dosma import file_constants as fc
from dosma.core import quant_vals as qv
from dosma.core.fitting import MonoExponentialFit
from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.format_io import ImageDataFormat
from dosma.core.io.nifti_io import NiftiReader
from dosma.core.med_volume import MedicalVolume
from dosma.core.quant_vals import QuantitativeValueType
from dosma.defaults import preferences
from dosma.scan_sequences.scans import ScanSequence
from dosma.tissues.tissue import Tissue
from dosma.utils import io_utils
from dosma.utils.cmd_line_utils import ActionWrapper

__all__ = ["Mapss"]

__EXPECTED_NUM_ECHO_TIMES__ = 7

__INITIAL_T1_RHO_VAL__ = 70.0
__T1_RHO_LOWER_BOUND__ = 0
__T1_RHO_UPPER_BOUND__ = 500

__INITIAL_T2_VAL__ = 30.0
__T2_LOWER_BOUND__ = 0
__T2_UPPER_BOUND__ = 100

__DECIMAL_PRECISION__ = 3

_logger = logging.getLogger(__name__)


class Mapss(ScanSequence):
    """MAPSS MRI sequence.

    Magnetization‐prepared angle‐modulated partitioned k‐space spoiled gradient echo snapshots
    (3D MAPSS) is a spoiled gradient (SPGR) sequence that reduce specific absorption rate (SAR),
    increase SNR, and reduce the extent of retrospective correction of contaminating T1 effects.

    The MAPSS sequence can be used to estimate both T1𝜌 and T2 quantitative values.
    MAPSS scans must also be intra-registered to ensure alignment between all volumes
    acquired at different echos and spin-lock times. Intra-registration is performed
    automatically upon construction. :math:`T_2` and :`T_{1\\rho}` fitting is also supported.

    References:
        X Li, ET Han, RF Busse, S Majumdar. In vivo t1ρ mapping in
        cartilage using 3d magnetization-prepared angle-modulated partitioned k-space spoiled
        gradient echo snapshots (3d mapss). Magnetic Resonance in Medicine, 59(2):298–307 (2008).
    """

    NAME = "mapss"

    def __init__(self, volumes: Sequence[MedicalVolume], echo_times: Sequence[float] = None):
        if not isinstance(volumes, Sequence):
            raise ValueError("`volumes` must be sequence of MedicalVolumes.")

        super().__init__(volumes)

        if echo_times is None:
            try:
                if all(x.headers() is not None for x in self.volumes):
                    echo_times = [x.get_metadata("EchoTime", float) for x in self.volumes]
            except (KeyError, AttributeError, RuntimeError) as e:
                raise ValueError(
                    f"Could not extract echo times from header. "
                    f"Please specify `echo_times` argument - {e}"
                )

        self.echo_times = echo_times

    def __validate_scan__(self):
        return len(self.volumes) == 7

    def __intraregister__(self, volumes: List[MedicalVolume]):
        """Intraregister volumes.

        Sets `self.volumes` to intraregistered volumes.

        Args:
            volumes (list[MedicalVolume]): Volumes to register.

        Raises:
            TypeError: If `volumes` is not `list[MedicalVolume]`.
        """
        if (
            (not volumes)
            or (type(volumes) is not list)
            or (len(volumes) != __EXPECTED_NUM_ECHO_TIMES__)
        ):
            raise TypeError("`volumes` must be of type List[MedicalVolume]")

        num_echos = len(volumes)

        _logger.info("")
        _logger.info("==" * 40)
        _logger.info("Intraregistering...")
        _logger.info("==" * 40)

        # temporarily save subvolumes as nifti file
        raw_volumes_base_path = io_utils.mkdirs(os.path.join(self.temp_path, "raw"))

        # Use first subvolume as a basis for registration
        # Save in nifti format to use with elastix/transformix
        volume_files = []
        for echo_index in range(num_echos):
            filepath = os.path.join(raw_volumes_base_path, "{:03d}.nii.gz".format(echo_index))
            volume_files.append(filepath)

            volumes[echo_index].save_volume(filepath, data_format=ImageDataFormat.nifti)

        target_echo_index = 0
        target_image_filepath = volume_files[target_echo_index]

        nr = NiftiReader()
        intraregistered_volumes = [deepcopy(volumes[target_echo_index])]
        for echo_index in range(1, num_echos):
            moving_image = volume_files[echo_index]

            reg = Registration()
            reg.inputs.fixed_image = target_image_filepath
            reg.inputs.moving_image = moving_image
            reg.inputs.output_path = io_utils.mkdirs(
                os.path.join(self.temp_path, "intraregistered", "{:03d}".format(echo_index))
            )
            reg.inputs.parameters = [fc.ELASTIX_AFFINE_PARAMS_FILE]
            reg.terminal_output = preferences.nipype_logging
            _logger.info("Registering {} -> {}".format(str(echo_index), str(target_echo_index)))
            tmp = reg.run()

            warped_file = tmp.outputs.warped_file
            intrareg_vol = nr.load(warped_file)

            # copy affine from original volume, because nifti changes loading accuracy
            intrareg_vol = MedicalVolume(
                volume=intrareg_vol.volume,
                affine=volumes[echo_index].affine,
                headers=deepcopy(volumes[echo_index].headers()),
            )

            intraregistered_volumes.append(intrareg_vol)

        self.volumes = intraregistered_volumes

    def intraregister(self):
        """Intra-register volumes."""
        self.__intraregister__(self.volumes)

    def generate_t1_rho_map(
        self, tissue: Tissue = None, mask_path: str = None, num_workers: int = 0
    ):
        """
        Generate 3D T1-rho map and r-squared fit map using mono-exponential fit
        across subvolumes acquired at different echo times.

        Args:
            tissue (Tissue): Tissue to generate quantitative value for.
            mask_path (str): File path to mask of ROI to analyze
            num_workers (int, optional): Number of subprocesses to use for fitting.
                If `0`, will execute on the main thread.

        Returns:
            qv.T1Rho: T1-rho fit for tissue.
        """
        echo_inds = range(4)
        bounds = (__T1_RHO_LOWER_BOUND__, __T1_RHO_UPPER_BOUND__)
        tc0 = "polyfit"
        decimal_precision = __DECIMAL_PRECISION__

        qv_map = self.__fitting_helper(
            qv.T1Rho, echo_inds, tissue, bounds, tc0, decimal_precision, mask_path, num_workers
        )

        return qv_map

    def generate_t2_map(self, tissue: Tissue = None, mask_path: str = None, num_workers: int = 0):
        """
        Generate 3D T2 map and r-squared fit map using mono-exponential fit
        across subvolumes acquired at different echo times.

        Args:
            tissue (Tissue): Tissue to generate quantitative value for.
            mask_path (str): File path to mask of ROI to analyze
            num_workers (int, optional): Number of subprocesses to use for fitting.
                If `0`, will execute on the main thread.

        Returns:
            qv.T2: T2 fit for tissue.
        """
        echo_inds = [0, 4, 5, 6]
        bounds = (__T2_LOWER_BOUND__, __T2_UPPER_BOUND__)
        tc0 = "polyfit"
        decimal_precision = __DECIMAL_PRECISION__

        qv_map = self.__fitting_helper(
            qv.T2, echo_inds, tissue, bounds, tc0, decimal_precision, mask_path, num_workers
        )

        return qv_map

    def __fitting_helper(
        self,
        qv_type: QuantitativeValueType,
        echo_inds: Sequence[int],
        tissue: Tissue,
        bounds,
        tc0,
        decimal_precision,
        mask_path,
        num_workers,
    ):
        echo_info = [(self.echo_times[i], self.volumes[i]) for i in echo_inds]

        # sort by echo time
        echo_info = sorted(echo_info, key=lambda x: x[0])

        xs = [et for et, _ in echo_info]
        ys = [vol for _, vol in echo_info]

        # only calculate for focused region if a mask is available, this speeds up computation
        mask = tissue.get_mask()
        if mask_path is not None:
            mask = (
                fio_utils.generic_load(mask_path, expected_num_volumes=1)
                if isinstance(mask_path, (str, os.PathLike))
                else mask_path
            )

        mef = MonoExponentialFit(
            bounds=bounds,
            tc0=tc0,
            decimal_precision=decimal_precision,
            num_workers=num_workers,
            verbose=True,
        )
        qv_map, r2 = mef.fit(xs, ys, mask=mask)

        quant_val_map = qv_type(qv_map)
        quant_val_map.add_additional_volume("r2", r2)

        tissue.add_quantitative_value(quant_val_map)

        return quant_val_map

    def _save(self, metadata, save_dir, fname_fmt=None, **kwargs):
        default_fmt = {MedicalVolume: "echo-{}"}
        default_fmt.update(fname_fmt if fname_fmt else {})
        return super()._save(metadata, save_dir, fname_fmt=default_fmt, **kwargs)

    @classmethod
    def cmd_line_actions(cls):
        """
        Provide command line information (such as name, help strings, etc)
        as list of dictionary.
        """
        intraregister_action = ActionWrapper(
            name=cls.intraregister.__name__, help="register volumes within this scan"
        )
        generate_t1_rho_map_action = ActionWrapper(
            name=cls.generate_t1_rho_map.__name__,
            aliases=["t1_rho"],
            param_help={
                "mask_path": (
                    "mask filepath (.nii.gz) to reduce computational time for fitting. "
                    "Not required if loading data (ie. `--l` flag) for tissue with mask."
                )
            },
            alternative_param_names={"mask_path": ["mask", "mp"]},
            help="generate T1-rho map using mono-exponential fitting",
        )
        generate_t2_map_action = ActionWrapper(
            name=cls.generate_t2_map.__name__,
            aliases=["t2"],
            param_help={
                "mask_path": (
                    "mask filepath (.nii.gz) to reduce computational time for fitting. "
                    "Not required if loading data (ie. `--l` flag) for tissue with mask."
                )
            },
            alternative_param_names={"mask_path": ["mask", "mp"]},
            help="generate T2 map using mono-exponential fitting",
        )

        return [
            (cls.intraregister, intraregister_action),
            (cls.generate_t1_rho_map, generate_t1_rho_map_action),
            (cls.generate_t2_map, generate_t2_map_action),
        ]



import itertools

import numpy as np
import seaborn as sns

from dosma import defaults

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

__all__ = ["downsample_slice", "write_regions"]


def downsample_slice(img_array, ds_factor, is_mask=False):
    """
    Takes in a 3D array and then downsamples in the z-direction by a
    user-specified downsampling factor.

    Args:
        img_array (np.ndarray): 3D numpy array for now (xres x yres x zres)
        ds_factor (int): Downsampling factor
        is_mask (:obj:`bool`, optional): If ``True``, ``img_array`` is a mask and will be binarized
            after downsampling. Defaults to `False`.

    Returns:
        np.ndarray: 3D numpy array of dimensions (xres x yres x zres//ds_factor)

    Examples:
        >>> input_image  = numpy.random.rand(4,4,4)
        >>> input_mask   = (a > 0.5) * 1.0
        >>> output_image = downsample_slice(input_mask, ds_factor = 2, is_mask = False)
        >>> output_mask  = downsample_slice(input_mask, ds_factor = 2, is_mask = True)
    """

    img_array = np.transpose(img_array, (2, 0, 1))
    L = list(img_array)

    def grouper(iterable, n):
        args = [iter(iterable)] * n
        return itertools.zip_longest(fillvalue=0, *args)

    final = np.array([sum(x) for x in grouper(L, ds_factor)])
    final = np.transpose(final, (1, 2, 0))

    # Binarize if it is a mask.
    if is_mask is True:
        final = (final >= 1) * 1

    return final


def write_regions(file_path, arr, plt_dict=None):
    """Write 2D array to region image where colors correspond to the region.

    All finite values should be >= 1.
    nan/inf value are ignored - written as white.

    Args:
        file_path (str): File path to save image.
        arr (np.ndarray): The 2D numpy array to convert to region image.
            Unique non-zero values correspond to different regions.
            Values that are `0` or `np.nan` will be written as white pixels.
        plt_dict (:obj:`dict`, optional): Dictionary of values to use when plotting with
            ``matplotlib.pyplot``. Keys are strings like `xlabel`, `ylabel`, etc.
            Use Key `labels` to specify a mapping from unique non-zero values in the array
            to names for the legend.
    """

    if len(arr.shape) != 2:
        raise ValueError("`arr` must be a 2D numpy array")

    unique_vals = np.unique(arr.flatten())
    if 0 in unique_vals:
        raise ValueError("All finite values in `arr` must be >=1")

    unique_vals = unique_vals[np.isfinite(unique_vals)]
    num_unique_vals = len(unique_vals)

    plt_dict_int = {"xlabel": "", "ylabel": "", "title": "", "labels": None}
    if plt_dict:
        plt_dict_int.update(plt_dict)
    plt_dict = plt_dict_int

    labels = plt_dict["labels"]
    if labels is None:
        labels = list(unique_vals)

    if len(labels) != num_unique_vals:
        raise ValueError(
            "len(labels) != num_unique_vals - %d != %d" % (len(labels), num_unique_vals)
        )

    cpal = sns.color_palette("pastel", num_unique_vals)

    arr_c = np.array(arr)
    arr_c = np.nan_to_num(arr_c)
    arr_c[arr_c > np.max(unique_vals)] = 0
    arr_rgb = np.ones([arr_c.shape[0], arr_c.shape[1], 3])

    plt.figure()
    plt.clf()
    custom_lines = []
    for i in range(num_unique_vals):
        unique_val = unique_vals[i]
        i0, i1 = np.where(arr_c == unique_val)
        arr_rgb[i0, i1, ...] = np.asarray(cpal[i])

        custom_lines.append(
            Line2D([], [], color=cpal[i], marker="o", linestyle="None", markersize=5)
        )

    plt.xlabel(plt_dict["xlabel"])
    plt.ylabel(plt_dict["ylabel"])
    plt.title(plt_dict["title"])

    lgd = plt.legend(
        custom_lines,
        labels,
        loc="upper center",
        bbox_to_anchor=(0.5, -defaults.DEFAULT_TEXT_SPACING),
        fancybox=True,
        shadow=True,
        ncol=3,
    )
    plt.imshow(arr_rgb)

    plt.savefig(file_path, bbox_extra_artists=(lgd,), bbox_inches="tight")



"""
DICOM I/O.

This module contains DICOM input/output helpers.

Note:

    1. Dicom utilizes LPS convention:
        - LPS: right --> left, anterior --> posterior, inferior --> superior
        - we will call it LPS+, such that letters correspond to increasing end of axis

Attributes:
    TOTAL_NUM_ECHOS_KEY (tuple[int]): Hexadecimal encoding of DICOM tag corresponding
        to number of echos.
"""

import copy
import functools
import itertools
import multiprocessing as mp
import os
import re
from math import ceil, log10
from typing import Collection, List, Sequence, Tuple, Union

import nibabel as nib
import numpy as np
import pydicom
from natsort import index_natsorted, natsorted
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map

from dosma.core import orientation as stdo
from dosma.core.io.format_io import DataReader, DataWriter, ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import AFFINE_DECIMAL_PRECISION, SCANNER_ORIGIN_DECIMAL_PRECISION

__all__ = ["DicomReader", "DicomWriter"]

TOTAL_NUM_ECHOS_KEY = (0x19, 0x107E)
PATH_LIKE = (str, os.PathLike)


class DicomReader(DataReader):
    """A class for reading DICOM files.

    Attributes:
        num_workers (int, optional): Number of workers to use for loading.
        verbose (bool, optional): If ``True``, show loading progress bar.
        group_by (str(s) or int(s), optional): DICOM attribute(s) used
            to group dicoms. This can be the attribute tag name (str) or tag
            number (int).
        sort_by (str(s) or int(s), optional): DICOM attribute(s) used
            to sort dicoms. This sorting is done after sorting files in alphabetical
            order.
        ignore_ext (bool, optional): If ``True``, ignore extension (``".dcm"``)
            when loading dicoms from directory.
        default_ornt (Tuple[str, str], optional): Default in-plane orientation to use if
            orientation cannot be determined from DICOM header. If not specified
            and orientation cannot be determined, error will be raised.
        data_format_code (ImageDataFormat): The supported image data format.

    Examples:
        >>> # Load single dicom
        >>> dr = DicomReader()
        >>> mv = dr.load("/path/to/dicom/file", group_by=None)[0]

        >>> # Load multi-echo MRI data
        >>> dr = DicomReader(num_workers=0, verbose=True)
        >>> mvs = dr.load("/dicoms/directory", group_by="EchoTime", sort_by="InstanceNumber")

        >>> # Use the same loader for multiple multi-echo time-series MRI scans
        >>> dr = DicomReader(group_by=["EchoTime", "TriggerTime", sort_by="InstanceNumber")
        >>> scans = [dr.load(dcm_dir) for dcm_dir in ["/dicom/dir1", "/dicom/dir2", "/dicom/dir3"]]
    """

    data_format_code = ImageDataFormat.dicom

    def __init__(
        self,
        num_workers: int = 0,
        verbose: bool = False,
        group_by: Union[str, int, Sequence[Union[str, int]]] = "EchoNumbers",
        sort_by: Union[str, int, Sequence[Union[str, int]]] = None,
        ignore_ext: bool = False,
        default_ornt: Tuple[str, str] = None,
    ):
        """
        Args:
            num_workers (int, optional): Number of workers to use for loading.
            verbose (bool, optional): If ``True``, show loading progress bar.
            group_by (:obj:`str(s)` or :obj:`int(s)`, optional): DICOM attribute(s) used
                to group dicoms. This can be the attribute tag name (str) or tag
                number (int).
            sort_by (:obj:`str(s)` or :obj:`int(s)`, optional): DICOM attribute(s) used
                to sort dicoms. This sorting is done after sorting files in alphabetical
                order.
            ignore_ext (bool, optional): If ``True``, ignore extension (``".dcm"``)
                when loading dicoms from directory.
            default_ornt (Tuple[str, str], optional): Default in-plane orientation to use if
                orientation cannot be determined from DICOM header. If not specified
                and orientation cannot be determined, error will be raised.
        """
        self.num_workers = num_workers
        self.verbose = verbose
        self.group_by = group_by
        self.sort_by = sort_by
        self.ignore_ext = ignore_ext
        self.default_ornt = default_ornt

    def get_files(
        self,
        path,
        include: Union[str, Sequence[str]] = None,
        exclude: Union[str, Sequence[str]] = None,
        ignore_hidden: bool = True,
        ignore_ext: bool = np._NoValue,
    ):
        """Get dicom files from directory.

        Args:
            path (str): Directory with dicom file(s).
            include (str(s), optional): Regex pattern(s) of files to include.
                Pattern matching will only be applied to the base file name.
                Used with :func:`re.match`.
            exclude (str(s), optional): Regex pattern(s) of files to exclude.
                Pattern matching will only be applied to the base file name.
                Used with :func:`re.match`.
            ignore_ext (bool, optional): If ``True``, ignore extension (`.dcm`)
                when loading dicoms from directory.
            ignore_hidden (bool, optional): If ``True``, ignores hidden
                files (files starting with ``"."``). Defaults to ``self.ignore_ext``.

        Returns:
            List[str]: Dicom file paths (in natsort order)

        Raises:
            NotADirectoryError: If ``path`` does not correspond to directory.
        """
        if not os.path.isdir(path):
            raise NotADirectoryError("`path` must be path to directory with dicoms.")

        ignore_ext = ignore_ext if ignore_ext != np._NoValue else self.ignore_ext

        include = _wrap_as_tuple(include, default=())
        exclude = _wrap_as_tuple(exclude, default=())
        assert isinstance(include, tuple)
        assert isinstance(exclude, tuple)
        if ignore_hidden:
            exclude += ("^\.",)  # noqa: W605

        possible_files = os.listdir(path)
        lstFilesDCM = []
        for f in possible_files:
            # If ignore extension, don't look for '.dcm' extension.
            is_file = os.path.isfile(os.path.join(path, f))
            match_ext = ignore_ext or self.data_format_code.is_filetype(f)
            if (
                is_file
                and match_ext
                and (not include or any(re.match(x, f) for x in include))
                and (not exclude or all(not re.match(x, f) for x in exclude))
            ):
                lstFilesDCM.append(os.path.join(path, f))

        lstFilesDCM = natsorted(lstFilesDCM)
        return lstFilesDCM

    def _handle_files(self, path, ignore_ext):
        """Gets and organize pydicom data from file(s) or directory.

        Args:
            path (str | path-like): The file path(s) or directory path to
                load from.
            ignore_ext (bool): If ``True``, ignore extension (``".dcm"``)
                when loading dicoms from directory.

        Returns:
            Sequence[path-like]: The filepaths for different slices.
        """
        if isinstance(path, str) or not isinstance(path, Sequence):
            if os.path.isdir(path):
                lstFilesDCM = self.get_files(path, ignore_hidden=True, ignore_ext=ignore_ext)
            elif os.path.isfile(path):
                lstFilesDCM = [path]
            else:
                raise IOError(f"No directory or file found - {path}")
        else:
            not_files = [x for x in path if not os.path.isfile(x)]
            if len(not_files) > 0:
                raise IOError(
                    "Files not found:\n{}".format("".join("\t{}\n".format(x) for x in not_files))
                )
            lstFilesDCM = path

        lstFilesDCM = natsorted(lstFilesDCM)
        if len(lstFilesDCM) == 0:
            raise FileNotFoundError("No valid dicom files found in {}".format(path))

        return lstFilesDCM

    def load(
        self,
        path_or_bytes: Union[str, os.PathLike, bytes, Sequence[Union[str, os.PathLike, bytes]]],
        group_by: Union[str, int, Sequence[Union[str, int]]] = np._NoValue,
        sort_by: Union[str, int, Sequence[Union[str, int]]] = np._NoValue,
        ignore_ext: bool = np._NoValue,
        default_ornt: Tuple[str, str] = np._NoValue,
    ):
        """Load dicoms into ``MedicalVolume``s grouped by ``group_by`` tag(s).

        When loading files from a directory, all hidden files (files starting with ``"."``)
        are ignored. Files are initially sorted in alphabetical order and subsequently by
        ``sort_by`` if specified.

        Args:
            path_or_bytes (`str(s)`): Directory with dicom files or dicom file(s).
                Dicom file(s) can either be the path or the bytes from the opened file.
            group_by (:obj:`str(s)` or :obj:`int(s)`, optional): DICOM attribute(s) used
                to group dicoms. This can be the attribute tag name (str) or tag
                number (int). Defaults to ``self.group_by``.
            sort_by (:obj:`str(s)` or :obj:`int(s)`, optional): DICOM attribute(s) used
                to sort dicoms. This sorting is done after sorting files in alphabetical
                order. Defaults to ``self.sort_by``.
            ignore_ext (bool, optional): If ``True``, ignore extension (``".dcm"``)
                when loading dicoms from directory. Defaults to ``self.ignore_ext``.
            default_ornt (Tuple[str, str], optional): Default in-plane orientation to use if
                orientation cannot be determined from DICOM header. If not specified
                and orientation cannot be determined, error will be raised.
                Defaults to ``self.default_ornt``.

        Returns:
            list[MedicalVolume]: Different volumes grouped by the `group_by` DICOM tag.

        Raises:
            ValueError: If `group_by` not specified or if single dicom file specified.
            IOError: If directory or dicom file(s) specified by `path` do not exist.
            FileNotFoundError: If no valid dicom files found.

        Note:
            Not setting the ``group_by`` argument could result in ill-formatted volumes.
            For best performance, specify ``group_by`` based on the attribute(s) differentiating
            different volumes in the scan.
        """
        group_by = group_by if group_by != np._NoValue else self.group_by
        sort_by = sort_by if sort_by != np._NoValue else self.sort_by
        ignore_ext = ignore_ext if ignore_ext != np._NoValue else self.ignore_ext
        default_ornt = default_ornt if default_ornt != np._NoValue else self.default_ornt

        group_by = _wrap_as_tuple(group_by, default=())
        sort_by = _wrap_as_tuple(sort_by, default=())

        if isinstance(path_or_bytes, PATH_LIKE) or (
            isinstance(path_or_bytes, Sequence) and isinstance(path_or_bytes[0], PATH_LIKE)
        ):
            lstFilesDCM = self._handle_files(path_or_bytes, ignore_ext)
        else:
            # We explicitly specify list/tuple because other objects can be sequences
            # (e.g. bytes) that we cannot thoroughly account for.
            lstFilesDCM = (
                [path_or_bytes] if not isinstance(path_or_bytes, (list, tuple)) else path_or_bytes
            )

        if self.num_workers:
            fn = functools.partial(pydicom.read_file, force=True)
            if self.verbose:
                dicom_slices = process_map(fn, lstFilesDCM, max_workers=self.num_workers)
            else:
                with mp.Pool(self.num_workers) as p:
                    dicom_slices = p.map(fn, lstFilesDCM)
        else:
            dicom_slices = [
                pydicom.read_file(fp, force=True)
                for fp in tqdm(lstFilesDCM, disable=not self.verbose)
            ]

        # Check if dicom file has the group_by element specified
        temp_dicom = dicom_slices[0]
        for _group in group_by:
            if _group not in temp_dicom:
                raise KeyError("Tag {} does not exist in dicom".format(_group))

        if sort_by:
            try:
                dicom_slices = natsorted(
                    dicom_slices,
                    key=lambda x: tuple(
                        _unpack_dicom_attr(x, attr, required=True) for attr in sort_by
                    ),
                )
            except KeyError as e:
                raise KeyError(f"Tag not found in dicom - {e}")

        dicom_data = {}
        for ds in dicom_slices:
            val_groupby = tuple(_unpack_dicom_attr(ds, attr, required=True) for attr in group_by)
            if val_groupby not in dicom_data.keys():
                dicom_data[val_groupby] = {"headers": [], "arr": []}

            dicom_data[val_groupby]["headers"].append(ds)
            dicom_data[val_groupby]["arr"].append(ds.pixel_array)

        vols = []
        for k in sorted(dicom_data.keys()):
            dd = dicom_data[k]
            headers = dd["headers"]
            if len(headers) == 0:
                continue
            arr = np.stack(dd["arr"], axis=-1)

            affine = to_RAS_affine(headers, default_ornt=default_ornt)

            vol = MedicalVolume(arr, affine, headers=headers)
            vols.append(vol)

        return vols

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    read = load  # pragma: no cover


class DicomWriter(DataWriter):
    """A class for writing volumes in DICOM format.

    Attributes:
        num_workers (int, optional): Number of workers to use for writing.
        verbose (bool, optional): If ``True``, show writing progress bar.
        fname_fmt (str, optional): Formatting string for filenames.
        sort_by (:obj:`str(s)` or :obj:`int(s)`, optional): DICOM attribute(s) used
            to define ordering of slices prior to writing. If not specified, this ordering
            will be defined by the order of blocks in ``volume``.
        data_format_code (ImageDataFormat): The supported image data format.

    Examples:
        >>> # Save MedicalVolume mv
        >>> dw = DicomWriter()
        >>> dw.save(mv, "/path/to/save/folder")

        >>> dw = DicomWriter(fname_fmt="I%05d.dcm", sort_by="InstanceNumber")
        >>> dw.save(mv, "/path/to/save/folder")
    """

    data_format_code = ImageDataFormat.dicom

    def __init__(
        self,
        num_workers: int = 0,
        verbose: bool = False,
        fname_fmt: str = None,
        sort_by: Union[str, int, Sequence[Union[str, int]]] = None,
    ):
        """

        Args:
            num_workers (int, optional): Number of workers to use for writing.
            verbose (bool, optional): If ``True``, show writing progress bar.
            fname_fmt (str, optional): Formatting string for filenames. Must contain ``%d``,
                which correspopnds to slice number. Defaults to
                ``"I%0{max(4, ceil(log10(num_slices)))}d.dcm"`` (e.g. ``"I0001.dcm"``).
            sort_by (str(s) or int(s), optional): DICOM attribute(s) used
                to define ordering of slices prior to writing. If not specified, this ordering
                will be defined by the order of blocks in ``volume``.
        """
        self.num_workers = num_workers
        self.verbose = verbose
        self.fname_fmt = fname_fmt
        self.sort_by = sort_by

    def save(
        self,
        volume: MedicalVolume,
        dir_path: str,
        fname_fmt: str = np._NoValue,
        sort_by: Union[str, int, Sequence[Union[str, int]]] = np._NoValue,
    ):
        """Save `medical volume` in dicom format.

        This function assumes headers for the volume (``volume.headers()``) exist
        for one spatial dimension. Headers for non-spatial dimensions are optional, but
        highly recommended. If provided, they will be used to write the volume. If not,
        headers will be appropriately broadcast to these dimensions. Note, this means
        that multiple files will have the same header information and will not be able
        to be loaded automatically.

        Currently header spatial information (orientation, origin, slicing between spaces,
        etc.) is not overwritten nor validated. All data must correspond to the same
        spatial information as specified in the headers to produce valid DICOM files.

        Args:
            volume (MedicalVolume): Volume to save.
            dir_path: Directory path to store dicom files. Dicoms are stored in directories,
                as multiple files are needed to store the volume.
            fname_fmt (str, optional): Formatting string for filenames. Must contain ``%d``,
                which correspopnds to slice number. Defaults to ``self.fname_fmt``.
            sort_by (``str``(s) or ``int``(s), optional): DICOM attribute(s) used
                to define ordering of slices prior to writing. If ``None``, this ordering
                will be defined by the order of blocks in ``volume``. Defaults to
                ``self.sort_by``.

        Raises:
            ValueError: If `im` does not have initialized headers. Or if `im` was flipped across
                any axis. Flipping changes scanner origin, which is currently not handled.
        """
        fname_fmt = fname_fmt if fname_fmt != np._NoValue else self.fname_fmt
        sort_by = sort_by if sort_by != np._NoValue else self.sort_by

        # Get orientation indicated by headers.
        headers = volume.headers()
        if headers is None:
            raise ValueError("MedicalVolume headers must be initialized to save as a dicom")

        sort_by = _wrap_as_tuple(sort_by, default=())

        # Reformat to put headers in last dimensions.
        single_dim = []
        full_dim = []
        for i, dim in enumerate(headers.shape[:3]):
            if dim == 1:
                single_dim.append(i)
            else:
                full_dim.append(i)
        if len(full_dim) > 1:
            raise ValueError(
                f"Only one spatial dimension can have headers. Got {len(full_dim)} - "
                f"headers.shape={headers.shape[:3]}"
            )
        new_orientation = (volume.orientation[x] for x in single_dim + full_dim)

        volume = volume.reformat(new_orientation)
        assert volume.headers().shape[:3] == (1, 1, volume.shape[2])

        # Reformat medical volume to expected orientation specified by dicom headers.
        # NOTE: This is temporary. Future fixes will allow us to modify header
        # data to match affine matrix.
        if len(volume.shape) > 3:
            shape = volume.shape[3:]
            multi_volumes = np.empty(shape, dtype=object)
            for dims in itertools.product(*[list(range(0, x)) for x in multi_volumes.shape]):
                multi_volumes[dims] = _format_volume_to_header(volume[(Ellipsis,) + dims])
            multi_volumes = multi_volumes.flatten()
            volume_arr = np.concatenate([v.volume for v in multi_volumes], axis=-1)
            headers = np.concatenate([v.headers(flatten=True) for v in multi_volumes], axis=-1)
        else:
            volume = _format_volume_to_header(volume)
            volume_arr = volume.volume
            headers = volume.headers(flatten=True)

        assert headers.ndim == 1
        assert volume_arr.shape[2] == len(
            headers
        ), "Dimension mismatch - {:d} slices but {:d} headers".format(
            volume_arr.shape[-1], len(headers)
        )

        if sort_by:
            idxs = np.asarray(
                index_natsorted(
                    headers,
                    key=lambda h: tuple(_unpack_dicom_attr(h, k, required=True) for k in sort_by),
                )
            )
            headers = headers[idxs]
            volume_arr = volume_arr[..., idxs]

        # Check if dir_path exists.
        os.makedirs(dir_path, exist_ok=True)

        num_slices = len(headers)
        if not fname_fmt:
            filename_format = "I%0" + str(max(4, ceil(log10(num_slices)))) + "d.dcm"
        else:
            filename_format = fname_fmt

        filepaths = [os.path.join(dir_path, filename_format % (s + 1)) for s in range(num_slices)]
        if self.num_workers:
            slices = [volume_arr[..., s] for s in range(num_slices)]
            if self.verbose:
                process_map(_write_dicom_file, slices, headers, filepaths)
            else:
                with mp.Pool(self.num_workers) as p:
                    out = p.starmap_async(_write_dicom_file, zip(slices, headers, filepaths))
                    out.wait()
        else:
            for s in tqdm(range(num_slices), disable=not self.verbose):
                _write_dicom_file(volume_arr[..., s], headers[s], filepaths[s])

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    write = save  # pragma: no cover


def to_RAS_affine(headers: List[pydicom.FileDataset], default_ornt: Tuple[str, str] = None):
    """Convert from LPS+ orientation (default for DICOM) to RAS+ standardized orientation.

    Args:
        headers (list[pydicom.FileDataset]): Headers for DICOM files to reorient.
            Files should correspond to single volume.

    Returns:
        np.ndarray: Affine matrix.
    """
    try:
        im_dir = headers[0].ImageOrientationPatient
    except AttributeError:
        im_dir = _decode_inplane_direction(headers, default_ornt=default_ornt)
        if im_dir is None:
            raise RuntimeError("Could not determine in-plane directions from headers.")
    try:
        in_plane_pixel_spacing = headers[0].PixelSpacing
    except AttributeError:
        try:
            in_plane_pixel_spacing = headers[0].ImagerPixelSpacing
        except AttributeError:
            raise RuntimeError(
                "Could not determine in-plane pixel spacing from headers. "
                "Neither attribute 'PixelSpacing' nor 'ImagerPixelSpacing' found."
            )

    orientation = np.zeros([3, 3])

    # Determine vector for in-plane pixel directions (i, j).
    i_vec, j_vec = (
        np.asarray(im_dir[:3]).astype(np.float64),
        np.asarray(im_dir[3:]).astype(np.float64),
    )  # unique to pydicom, please revise if using different library to load dicoms
    i_vec, j_vec = (
        np.round(i_vec, AFFINE_DECIMAL_PRECISION),
        np.round(j_vec, AFFINE_DECIMAL_PRECISION),
    )
    i_vec = i_vec * in_plane_pixel_spacing[0]
    j_vec = j_vec * in_plane_pixel_spacing[1]

    # Determine vector for through-plane pixel direction (k).
    # Compute difference in patient position between consecutive headers.
    # This is the preferred method to determine the k vector.
    # If single header, take cross product between i/j vectors.
    # These actions are done to avoid rounding errors that might result from float subtraction.
    if len(headers) > 1:
        k_vec = np.asarray(headers[1].ImagePositionPatient).astype(np.float64) - np.asarray(
            headers[0].ImagePositionPatient
        ).astype(np.float64)
    else:
        slice_thickness = headers[0].get("SliceThickness", 1.0)
        i_norm = 1 / np.linalg.norm(i_vec) * i_vec
        j_norm = 1 / np.linalg.norm(j_vec) * j_vec
        k_norm = np.cross(i_norm, j_norm)
        k_vec = k_norm / np.linalg.norm(k_norm) * slice_thickness
        if hasattr(headers[0], "SpacingBetweenSlices") and headers[0].SpacingBetweenSlices < 0:
            k_vec *= -1
    k_vec = np.round(k_vec, AFFINE_DECIMAL_PRECISION)

    orientation[:3, :3] = np.stack([j_vec, i_vec, k_vec], axis=1)
    scanner_origin = headers[0].get("ImagePositionPatient", np.zeros((3,)))
    scanner_origin = np.asarray(scanner_origin).astype(np.float64)
    scanner_origin = np.round(scanner_origin, SCANNER_ORIGIN_DECIMAL_PRECISION)

    affine = np.zeros([4, 4])
    affine[:3, :3] = orientation
    affine[:3, 3] = scanner_origin
    affine[:2, :] = -1 * affine[:2, :]
    affine[3, 3] = 1

    affine[affine == 0] = 0

    return affine


def _decode_inplane_direction(headers: Sequence[pydicom.FileDataset], default_ornt=None):
    """Helper function to decode in-plane direction from header(s).

    Recall the direction in dicoms are in cartesian order ``(x,y)``,
    but numpy/dosma are in matrix order ``(y,x)``. When adding new
    methods, make sure to account for this.

    Returns:
        np.ndarray: 6-element LPS direction array where first 3 elements define
            direction for x-direction (columns) and second 3 elements define
            direction for y-direction (rows)
    """
    _patient_ornt_to_nib = {"H": "S", "F": "I"}

    if (
        len(headers) == 1
        and hasattr(headers[0], "PatientOrientation")
        and headers[0].PatientOrientation
    ):
        # Decoder: patient orientation.
        # Patient orientation is only decoded along principal direction (e.g. "FR" -> "F").
        ornt = [
            _patient_ornt_to_nib.get(k[:1], k[:1]) for k in headers[0].PatientOrientation
        ]  # (x,y)
        ornt = stdo.orientation_nib_to_standard(ornt)
        affine = stdo.to_affine(ornt)
        affine[:2, :] = -1 * affine[:2, :]
        return np.concatenate([affine[:3, 0], affine[:3, 1]], axis=0)

    if default_ornt:
        affine = stdo.to_affine(default_ornt)
        affine[:2, :] = -1 * affine[:2, :]
        return np.concatenate([affine[:3, 0], affine[:3, 1]], axis=0)

    return None


def _format_volume_to_header(volume: MedicalVolume) -> MedicalVolume:
    """Reformats the volume according to its header.

    Args:
        volume (MedicalVolume): The volume to reformat.
            Must be 3D and have headers of shape (1, 1, volume.shape[2]).

    Returns:
        MedicalVolume: The reformatted volume.
    """
    headers = volume.headers()
    assert headers.shape == (1, 1, volume.shape[2])

    affine = to_RAS_affine(headers.flatten())
    orientation = stdo.orientation_nib_to_standard(nib.aff2axcodes(affine))

    # Currently do not support mismatch in scanner_origin.
    if tuple(affine[:3, 3]) != volume.scanner_origin:
        raise ValueError(
            "Scanner origin mismatch. "
            "Currently we do not handle mismatch in scanner origin "
            "(i.e. cannot flip across axis)"
        )

    volume = volume.reformat(orientation)
    assert volume.headers().shape == (1, 1, volume.shape[2])
    return volume


def _write_dicom_file(np_slice: np.ndarray, header: pydicom.FileDataset, file_path: str):
    """Replace data in header with 2D numpy array and write to `file_path`.

    Args:
        np_slice (np.ndarray): 2D slice to encode in dicom file.
        header (pydicom.FileDataset): DICOM header.
        file_path: File path to write to.
    """
    # Deep copy required in case headers are shared.
    header = copy.deepcopy(header)
    expected_dimensions = header.Rows, header.Columns
    assert (
        np_slice.shape == expected_dimensions
    ), "In-plane dimension mismatch - expected shape {}, got {}".format(
        str(expected_dimensions), str(np_slice.shape)
    )

    np_slice_bytes = np_slice.tobytes()
    bit_depth = int(len(np_slice_bytes) / (np_slice.shape[0] * np_slice.shape[1]) * 8)
    if bit_depth != header.BitsAllocated:
        np_slice = _update_np_dtype(np_slice, header.BitsAllocated)
        np_slice_bytes = np_slice.tobytes()
        bit_depth = int(len(np_slice_bytes) / (np_slice.shape[0] * np_slice.shape[1]) * 8)

    assert bit_depth == header.BitsAllocated, "Bit depth mismatch: Expected {:d} got {:d}".format(
        header.BitsAllocated, bit_depth
    )

    header.PixelData = np_slice_bytes

    header.save_as(file_path)


def _update_np_dtype(arr: np.ndarray, bit_depth: int):
    """Create copy of np_array with bit-depth and type specified here.

    Note pydicom only supports writing dicoms with bit-depth 8/16 - only supports bit depths 8/16.

    Args:
        arr (np.ndarray): Numpy array to put into given bit depth.
        bit_depth (int): Bit depth for writing dicom. Must be either `8` or `16`.

    Returns:
        np.ndarray: Copy of input np_array.

    Raises:
        ValueError: If `arr` out of bit-depth range.
        TypeError: If `arr` contains float values out of supported float types.
    """
    assert bit_depth in [8, 16], "Only bit-depths of 8 and 16 are currently supported."
    dtype_dict = {
        8: [(np.int8, -128, 127), (np.uint8, 0, 255)],
        16: [
            (np.uint16, 0, 2 ** 16 - 1),
            (np.int16, -(2 ** 15), 2 ** 15),
            (np.float16, -6.55e4, 6.55e4 - 1),
        ],
    }
    supported_floats = [np.float16]
    curr_min = np.min(arr)
    curr_max = np.max(arr)
    contains_float = (arr % 1 != 0).any()

    dtypes = dtype_dict[bit_depth]

    new_dtype = None
    for dtype, dtype_min, dtype_max in dtypes:
        if curr_min < dtype_min or curr_max > dtype_max:
            continue
        new_dtype = dtype
        break
    if not new_dtype:
        raise ValueError(
            "Cannot cast numpy array ({}) to bit-depth of {} bits".format(str(arr.dtype), bit_depth)
        )

    if contains_float and new_dtype not in supported_floats:
        raise TypeError(
            "Array contains float. Cannot cast to numpy array ({}) to {}".format(
                str(arr.dtype), new_dtype
            )
        )

    return arr.astype(new_dtype)


def _unpack_dicom_attr(header, attr, required=False):
    if not required:
        val = header.get(attr)
    else:
        try:
            val = header[attr]
        except KeyError:
            raise KeyError(f"Tag {attr} missing from dicom")

    if type(val) is pydicom.DataElement:
        val = val.value
    return val


def _wrap_as_tuple(x, default=None):
    """Wraps individual values as tuple."""
    if default is not None and not x:
        return default

    if isinstance(x, str) or not isinstance(x, Sequence):
        x = (x,)
    elif isinstance(x, Sequence) and not isinstance(x, tuple):
        x = tuple(x)
    return x



import os
import sys

import h5py
import nibabel
import nipype
import numpy as np
import pandas as pd
import pydicom
import scipy as sp
import skimage
from tabulate import tabulate

__all__ = ["collect_env_info"]


def collect_env_info():
    """Collect environment information for reporting issues.

    Run this function when reporting issues on Github.
    """
    data = []
    data.append(("sys.platform", sys.platform))
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    try:
        import dosma  # noqa

        data.append(("dosma", dosma.__version__ + " @" + os.path.dirname(dosma.__file__)))
    except ImportError:
        data.append(("dosma", "failed to import"))

    # Required packages
    data.append(("h5py", h5py.__version__))
    data.append(("nibabel", nibabel.__version__))
    data.append(("nipype", nipype.__version__))
    data.append(("pandas", pd.__version__))
    data.append(("pydicom", pydicom.__version__))
    data.append(("scipy", sp.__version__))
    data.append(("skimage", skimage.__version__))

    # Optional packages
    try:
        import cupy

        data.append(("cupy", cupy.__version__))
    except ImportError:
        pass

    try:
        import tensorflow

        data.append(("tensorflow", tensorflow.__version__))
    except ImportError:
        pass

    try:
        import keras

        data.append(("keras", keras.__version__))
    except ImportError:
        pass

    try:
        import SimpleITK as sitk

        data.append(("SimpleITK", sitk.__version__))
    except ImportError:
        pass

    try:
        import sigpy

        data.append(("sigpy", sigpy.__version__))
    except ImportError:
        pass

    env_str = tabulate(data)
    return env_str



import logging
import os
import sys
import tkinter as tk
from tkinter import IntVar, Radiobutton, filedialog, messagebox, ttk
from typing import Dict

import numpy as np
import Pmw
from skimage.color import label2rgb
from skimage.measure import label

from dosma.cli import SUPPORTED_QUANTITATIVE_VALUES, SUPPORTED_SCAN_TYPES, parse_args
from dosma.core.io import format_io_utils as fio_utils
from dosma.core.orientation import AXIAL, CORONAL, SAGITTAL
from dosma.gui.dosma_gui import ScanReader
from dosma.gui.gui_utils.filedialog_reader import FileDialogReader
from dosma.gui.im_viewer import IndexTracker
from dosma.gui.preferences_viewer import PreferencesManager
from dosma.msk import knee

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

matplotlib.use("TkAgg")
LARGE_FONT = ("Verdana", 12)

_logger = logging.getLogger(__name__)


class DosmaViewer(tk.Tk):
    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)

        container = tk.Frame(self)
        container.pack(side="top", fill="both", expand=True)
        container.grid_rowconfigure(0, weight=1)
        container.grid_columnconfigure(0, weight=1)

        self.frames = {}
        self.protocol("WM_DELETE_WINDOW", self.on_closing)
        for F in (StartPage, DosmaFrame, PageThree, AnalysisFrame):
            frame = F(container, self)

            self.frames[F] = frame

            frame.grid(row=0, column=0, sticky="nsew")

        self.show_frame(StartPage)

        self.pref = PreferencesManager()

    def on_closing(self):
        if messagebox.askokcancel("Quit", "Do you want to quit?"):
            sys.exit()

    def show_frame(self, cont):
        frame = self.frames[cont]
        frame.tkraise()

    def show_preferences(self):
        self.pref.show_window(self)


class StartPage(tk.Frame):
    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        # photo = tk.PhotoImage(file="./defaults/skel-rotate.gif")
        # label1 = tk.Label(image=photo)
        # label1.pack()

        label = tk.Label(self, text="Start Page", font=LARGE_FONT)
        label.pack(pady=10, padx=10)

        button2 = ttk.Button(self, text="Scan", command=lambda: controller.show_frame(DosmaFrame))
        button2.pack()

        button3 = ttk.Button(
            self, text="Knee Analysis", command=lambda: controller.show_frame(AnalysisFrame)
        )
        button3.pack()

        button3 = ttk.Button(
            self, text="Image Viewer", command=lambda: controller.show_frame(PageThree)
        )
        button3.pack()

        button3 = ttk.Button(
            self, text="Preferences", command=lambda: controller.show_preferences()
        )
        button3.pack()


class AnalysisFrame(tk.Frame):
    __TISSUES_KEY = "Tissues"
    __QUANTITATIVE_VALUES_KEY = "Quantitative values"
    __LOAD_PATH_KEY = "Load data"

    __PID_KEY = "pid"
    __MEDIAL_TO_LATERAL_ORIENTATION_KEY = "ml"

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)

        self.manager: Dict = {}
        self.gui_manager: Dict = {}
        self.balloon = Pmw.Balloon()

        self.__init_manager()

        self.__base_gui()
        self.preferences = PreferencesManager()
        self.file_dialog_reader = FileDialogReader()
        self.scan_reader = ScanReader(self)

        button1 = ttk.Button(self, text="Home", command=lambda: controller.show_frame(StartPage))
        button1.pack(anchor="se", side="right")

        button1 = ttk.Button(self, text="Run", command=lambda: self.execute())
        button1.pack(anchor="sw", side="left")

    def execute(self):
        try:
            load_path = self.manager[self.__LOAD_PATH_KEY].get()
            if not load_path:
                raise ValueError("Load path not defined")

            preferences_str = self.preferences.get_cmd_line_str().strip()

            tissue_str = ""
            for c, t in enumerate(self.manager[self.__TISSUES_KEY]):
                if t.get():
                    tissue_str += "--%s " % knee.SUPPORTED_TISSUES[c].STR_ID
            tissue_str = tissue_str.strip()

            if not tissue_str:
                raise ValueError("No tissues selected")

            qv_str = ""
            for c, qv in enumerate(self.manager[self.__QUANTITATIVE_VALUES_KEY]):
                if qv.get():
                    qv_str += "--%s " % SUPPORTED_QUANTITATIVE_VALUES[c].name.lower()
            qv_str = qv_str.strip()

            if not qv_str:
                raise ValueError("No quantitative values selected")

            pid = self.manager[self.__PID_KEY].get()
            medial_to_lateral = self.manager[self.__MEDIAL_TO_LATERAL_ORIENTATION_KEY].get()

            if not pid:
                raise ValueError("No PID was provided")

            # analysis string
            str_f = "--l %s %s knee %s --pid %s %s %s" % (
                load_path,
                preferences_str,
                tissue_str,
                pid,
                "--ml" if medial_to_lateral else "",
                qv_str,
            )
            str_f = str_f.strip()
            parse_args(str_f.split())
        except Exception as e:
            tk.messagebox.showerror(str(type(e)), e.__str__())

    def __init_manager(self):
        self.manager[self.__LOAD_PATH_KEY] = tk.StringVar()
        self.manager[self.__TISSUES_KEY] = [
            tk.BooleanVar() for i in range(len(knee.SUPPORTED_TISSUES))
        ]
        self.manager[self.__QUANTITATIVE_VALUES_KEY] = [
            tk.BooleanVar() for i in range(len(SUPPORTED_QUANTITATIVE_VALUES))
        ]

        self.manager[self.__PID_KEY] = tk.StringVar()
        self.manager[self.__MEDIAL_TO_LATERAL_ORIENTATION_KEY] = tk.BooleanVar()

    def __display_pid_info(self):
        hb = tk.Frame(self)
        hb.pack(side="top", anchor="nw")
        _label = tk.Label(hb, text=self.__PID_KEY.upper())
        _label.pack(side="left", anchor="w", pady=10)
        t = tk.Entry(hb, textvariable=self.manager[self.__PID_KEY])
        t.pack(side="left", anchor="w", pady=10)
        self.balloon.bind(_label, "Patient id")

    def __display_data_loader(self):
        hb = tk.Frame(self)

        filedialog = FileDialogReader(self.manager[self.__LOAD_PATH_KEY])
        b = tk.Button(
            hb,
            text=self.__LOAD_PATH_KEY,
            command=lambda fd=filedialog: self.manager[self.__LOAD_PATH_KEY].set(
                fd.get_save_dirpath()
            ),
        )
        b.pack(side="left", anchor="nw", pady=10)

        _label = tk.Label(hb, textvariable=self.manager[self.__LOAD_PATH_KEY])
        _label.pack(side="left", anchor="nw", pady=10)

        hb.pack(side="top", anchor="nw")

    def __display_multi_option(self, label, options_list, boolvar_list):
        hb = tk.Frame(self)
        _label = tk.Label(hb, text="%s:" % label)
        _label.pack(side="left", anchor="w")
        hb.pack(side="top", anchor="nw")
        frames = [tk.Frame(hb)] * (len(options_list) // 3 + 1)
        for ind, option in enumerate(options_list):
            f = frames[ind // 3]
            b = tk.Checkbutton(f, text=option, variable=boolvar_list[ind])
            b.pack(side="top", anchor="nw", pady=5)

        for f in frames:
            f.pack(side="left", anchor="nw")

        return hb

    def __display_tissues(self):
        tissue_names = [x.FULL_NAME for x in knee.SUPPORTED_TISSUES]
        _label = self.__display_multi_option(
            self.__TISSUES_KEY, tissue_names, self.manager[self.__TISSUES_KEY]
        )
        self.balloon.bind(_label, "Tissues to analyze")

    def __display_quant_vals(self):
        quantitative_value_names = [x.name for x in SUPPORTED_QUANTITATIVE_VALUES]
        _label = self.__display_multi_option(
            self.__QUANTITATIVE_VALUES_KEY,
            quantitative_value_names,
            self.manager[self.__QUANTITATIVE_VALUES_KEY],
        )
        self.balloon.bind(_label, "Quantitative values to analyze")

    def __display_knee_info(self):
        hb = tk.Frame(self)
        hb.pack(side="top", anchor="nw")
        _label = tk.Label(hb, text="Medial -> Lateral: ")
        _label.pack(side="left", anchor="w", pady=10)
        t = tk.Checkbutton(hb, variable=self.manager[self.__MEDIAL_TO_LATERAL_ORIENTATION_KEY])
        t.pack(side="left", anchor="w", pady=10)

        self.balloon.bind(_label, "Select if Dicoms proceed in medial->lateral direction")

    def __base_gui(self):
        self.__display_data_loader()
        self.__display_pid_info()
        self.__display_tissues()
        self.__display_knee_info()
        self.__display_quant_vals()


class DosmaFrame(tk.Frame):
    __SCAN_KEY = "Scan"
    __TISSUES_KEY = "Tissues"

    __DICOM_PATH_KEY = "Read dicoms"
    __LOAD_PATH_KEY = "Load data"

    __SAVE_PATH_KEY = "Save path"

    __DATA_KEY = "data"  # Track option menu for dicom/load path
    __DATA_PATH_KEY = "datapath"  # Track filepath associated with option menu

    __IGNORE_EXTENSION_KEY = "Ignore extension"

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)

        self.file_dialog_reader = FileDialogReader()

        self.manager: Dict = {}
        self.gui_manager: Dict = {}
        self.balloon = Pmw.Balloon()

        self.__init_manager()

        self.__base_gui()
        self.preferences = PreferencesManager()
        self.scan_reader = ScanReader(self)

        button1 = ttk.Button(self, text="Home", command=lambda: controller.show_frame(StartPage))
        button1.pack(anchor="se", side="right")

        button1 = ttk.Button(self, text="Run", command=lambda: self.execute())
        button1.pack(anchor="sw", side="left")

        self.InitUI()

    def execute(self):
        try:
            save_path = self.manager[self.__SAVE_PATH_KEY].get()
            if not save_path:
                raise ValueError("Save path not defined")

            action_str = self.scan_reader.get_cmd_line_str().strip()

            if not action_str:
                raise ValueError("No action selected")

            preferences_str = self.preferences.get_cmd_line_str().strip()

            source = "d"
            if self.manager[self.__DATA_KEY].get() == self.__LOAD_PATH_KEY:
                source = "l"

            tissue_str = ""
            for c, t in enumerate(self.manager[self.__TISSUES_KEY]):
                if t.get():
                    tissue_str += "--%s " % knee.SUPPORTED_TISSUES[c].STR_ID
            tissue_str = tissue_str.strip()

            if not tissue_str:
                raise ValueError("No tissues selected")

            ignore_ext = self.manager[self.__IGNORE_EXTENSION_KEY].get()

            str_f = "--%s %s --s %s %s %s %s %s %s" % (
                source,
                self.manager[self.__DATA_PATH_KEY].get(),
                save_path,
                preferences_str,
                "--ignore_ext" if ignore_ext else "",
                self.manager[self.__SCAN_KEY].get(),
                tissue_str,
                action_str,
            )

            _logger.info("CMD LINE INPUT: %s" % str_f)

            parse_args(str_f.split())
        except Exception as e:
            tk.messagebox.showerror(str(type(e)), e.__str__())

    def __init_manager(self):
        self.manager[self.__SCAN_KEY] = tk.StringVar()
        self.manager[self.__TISSUES_KEY] = [
            tk.BooleanVar() for i in range(len(knee.SUPPORTED_TISSUES))
        ]
        self.manager[self.__DATA_KEY] = tk.StringVar()
        self.manager[self.__DATA_PATH_KEY] = tk.StringVar()

        self.manager[self.__SCAN_KEY].trace_add("write", self.__on_scan_change)
        self.manager[self.__SAVE_PATH_KEY] = tk.StringVar()
        self.manager[self.__IGNORE_EXTENSION_KEY] = tk.BooleanVar()

    def __on_scan_change(self, *args):
        scan_id = self.manager[self.__SCAN_KEY].get()
        scan = None
        for x in SUPPORTED_SCAN_TYPES:
            if x.NAME == scan_id:
                scan = x

        self.scan_reader.load_scan(scan)

        assert scan is not None, "No scan selected"

    def __update_svar(self, *args):
        svar = self.manager[self.__DATA_PATH_KEY]
        selected_option = self.manager[self.__DATA_KEY].get()
        if selected_option == self.__DICOM_PATH_KEY:
            fp = self.file_dialog_reader.get_volume_filepath(
                selected_option, im_type=fio_utils.ImageDataFormat.dicom
            )
        elif selected_option == self.__LOAD_PATH_KEY:
            fp = self.file_dialog_reader.get_dirpath(selected_option)
        else:
            raise ValueError("%s key not found" % self.__DATA_KEY)

        if not fp:
            svar.set("")
            return

        svar.set(fp)

        if selected_option == self.__LOAD_PATH_KEY:
            self.manager[self.__SAVE_PATH_KEY].set(fp)

    def __display_data_loader(self):
        s_var = self.manager[self.__DATA_PATH_KEY]

        hb = tk.Frame(self)

        label = tk.Label(hb, text="Data source: ")
        label.pack(side="left", anchor="nw", pady=10)

        options = [self.__DICOM_PATH_KEY, self.__LOAD_PATH_KEY]
        menu = tk.OptionMenu(
            hb, self.manager[self.__DATA_KEY], *options, command=self.__update_svar
        )
        menu.pack(side="left", anchor="nw", pady=10)

        label = tk.Label(hb, textvariable=s_var)
        label.pack(side="left", anchor="nw", pady=10)

        hb.pack(side="top", anchor="nw")
        self.balloon.bind(hb, "Read dicoms or load data")

        hb = tk.Frame(self)

        # filedialog = FileDialogReader(self.manager[self.__SAVE_PATH_KEY])
        b = tk.Button(
            hb,
            text=self.__SAVE_PATH_KEY,
            command=lambda fd=self.file_dialog_reader: self.manager[self.__SAVE_PATH_KEY].set(
                fd.get_save_dirpath()
            ),
        )
        b.pack(side="left", anchor="nw", pady=10)

        label = tk.Label(hb, textvariable=self.manager[self.__SAVE_PATH_KEY])
        label.pack(side="left", anchor="nw", pady=10)

        hb.pack(side="top", anchor="nw")

        hb = tk.Frame(self)

        b = tk.Checkbutton(
            hb, text=self.__IGNORE_EXTENSION_KEY, variable=self.manager[self.__IGNORE_EXTENSION_KEY]
        )
        b.pack(side="left", anchor="nw", pady=10)
        self.balloon.bind(b, "Ignore '.dcm' extension when loading dicoms")

        hb.pack(side="top", anchor="nw")

    def __display_tissues(self):
        hb = tk.Frame(self)
        _label = tk.Label(hb, text="Tissues:")
        _label.pack(side="left", anchor="w")
        hb.pack(side="top", anchor="nw")
        frames = [tk.Frame(hb)] * (len(knee.SUPPORTED_TISSUES) // 3 + 1)
        for ind, tissue in enumerate(knee.SUPPORTED_TISSUES):
            f = frames[ind // 3]
            b = tk.Checkbutton(
                f, text=tissue.FULL_NAME, variable=self.manager[self.__TISSUES_KEY][ind]
            )
            b.pack(side="top", anchor="nw", pady=5)

        for f in frames:
            f.pack(side="left", anchor="nw")

        self.balloon.bind(_label, "Tissues to analyze")

    def __base_gui(self):
        self.__display_data_loader()
        self.__display_tissues()

        hb = tk.Frame(self)
        scan_label = tk.Label(hb, text="Scan:")
        scan_label.pack(side="left", anchor="nw", pady=10)
        options = [x.NAME for x in SUPPORTED_SCAN_TYPES]
        scan_dropdown = tk.OptionMenu(hb, self.manager[self.__SCAN_KEY], *options)
        scan_dropdown.pack(side="left", anchor="nw", pady=10)
        hb.pack(side="top", anchor="nw")

    def InitUI(self):
        self.text_box = tk.Text(self, wrap="word", height=11, width=50)
        self.text_box.pack(anchor="s", side="bottom")


class PageThree(tk.Frame):
    SUPPORTED_FORMATS = (("nifti files", "*.nii\.gz"), ("dicom files", "*.dcm"))  # noqa: W605
    __base_filepath = "../"

    _ORIENTATIONS = [("sagittal", SAGITTAL), ("coronal", CORONAL), ("axial", AXIAL)]

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self._im_display = None
        self.binding_vars: Dict = {}
        fig, ax = plt.subplots(1, 1)
        X = np.random.rand(20, 20, 40)

        self.tracker = IndexTracker(ax, X)

        canvas = FigureCanvasTkAgg(fig, self)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        canvas.mpl_connect("scroll_event", self.tracker.onscroll)

        toolbar = NavigationToolbar2Tk(canvas, self)
        toolbar.update()
        canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        self.im = None
        self.mask = None
        self._im_display = None

        button1 = ttk.Button(
            self, text="Back to Home", command=lambda: controller.show_frame(StartPage)
        )
        button1.pack(side=tk.BOTTOM, anchor="sw")

        button2 = ttk.Button(self, text="Load main image", command=self.load_volume_callback)
        button2.pack()

        button3 = ttk.Button(self, text="Load mask", command=self.load_mask_callback)
        button3.pack()

        self.init_reformat_display()

    def __reformat_callback(self, *args):
        self.im_update()

    def init_reformat_display(self):
        orientation_var = IntVar(0)
        orientation_var.trace_add("write", self.__reformat_callback)
        count = 0
        for text, _value in self._ORIENTATIONS:
            b = Radiobutton(self, text=text, variable=orientation_var, value=count)
            b.pack(side=tk.TOP, anchor="w")
            count += 1
        self._orientation = orientation_var

    def load_volume_callback(self):
        im = self.load_volume()
        if not im:
            return
        self.im = im
        self.mask = None

        self.im_update()

    def load_mask_callback(self):
        if not self.im:
            messagebox.showerror("Loading mask failed", "Main image must be loaded prior to mask")
            return

        mask = self.load_volume("Load mask")
        mask.reformat(self.im.orientation, inplace=True)
        try:
            self.__verify_mask_size(self.im.volume, mask.volume)
        except Exception as e:
            messagebox.showerror("Loading mask failed", str(e))
            return

        self.mask = mask
        self.im_update()

    def __verify_mask_size(self, im: np.ndarray, mask: np.ndarray):
        if mask.ndim != 3:
            raise ValueError("Dimension mismatch. Mask must be 3D")
        if im.shape != mask.shape:
            raise ValueError(
                "Dimension mismatch. Image of shape %s, but mask of shape %s"
                % (str(im.shape), str(mask.shape))
            )

    def im_update(self):
        orientation = self.orientation
        self.im.reformat(orientation, inplace=True)
        im = self.im.volume
        im = im / np.max(im)
        if self.mask:
            self.mask.reformat(orientation, inplace=True)
            label_image = label(self.mask.volume)
            im = self.__labeltorgb_3d__(im, label_image, 0.3)

        self.im_display = im

    def __labeltorgb_3d__(self, im: np.ndarray, labels: np.ndarray, alpha: float = 0.3):
        im_rgb = np.zeros(im.shape + (3,))  # rgb channel
        for s in range(im.shape[2]):
            im_slice = im[..., s]
            labels_slice = labels[..., s]
            im_rgb[..., s, :] = label2rgb(labels_slice, image=im_slice, bg_label=0, alpha=alpha)
        return im_rgb

    def load_volume(self, title="Select volume file(s)"):
        files = filedialog.askopenfilenames(initialdir=self.__base_filepath, title=title)
        if len(files) == 0:
            return

        filepath = files[0]
        self.__base_filepath = os.path.dirname(filepath)

        if filepath.endswith(".dcm"):
            filepath = os.path.dirname(filepath)

        im = fio_utils.generic_load(filepath, 1)

        return im

    @property
    def orientation(self):
        ind = self._orientation.get()
        return self._ORIENTATIONS[ind][1]

    @property
    def im_display(self):
        return self._im_display

    @im_display.setter
    def im_display(self, value):
        self._im_display = value
        self.tracker.x = self._im_display



import os
from tkinter import StringVar, filedialog

from dosma.core.io import format_io_utils as fio_utils
from dosma.gui.preferences_viewer import PreferencesManager


class FileDialogReader:
    SUPPORTED_FORMATS = (("nifti files", "*.nii\.gz"), ("dicom files", "*.dcm"))  # noqa: W605
    __base_filepath = "../"

    def __init__(self, string_var: StringVar = None):
        self.preferences = PreferencesManager()
        self.string_var = string_var

    def load_volume(self, title="Select volume file(s)"):
        filepath = self.get_volume_filepath(title)

        im = fio_utils.generic_load(filepath, 1)

        return im

    def get_volume_filepath(self, title="Select path", im_type: fio_utils.ImageDataFormat = None):
        filetypes = None
        if im_type is fio_utils.ImageDataFormat.dicom:
            filetypes = ((im_type.name, "*.dcm"),)

        files = filedialog.askopenfilenames(
            initialdir=self.__base_filepath, title=title, filetypes=filetypes
        )
        if len(files) == 0:
            return

        filepath = files[0]
        self.__base_filepath = os.path.dirname(filepath)

        if filepath.endswith(".dcm"):
            filepath = os.path.dirname(filepath)

        if self.string_var:
            self.string_var.set(filepath)

        return filepath

    def get_filepath(self, title="Select file"):
        file_str = filedialog.askopenfilename(initialdir=self.__base_filepath, title=title)
        if not file_str:
            return

        if self.string_var:
            self.string_var.set(file_str)

        return file_str

    def get_dirpath(self, title="Select directory"):
        file_str = filedialog.askdirectory(initialdir=self.__base_filepath, title=title)
        if not file_str:
            return

        if self.string_var:
            self.string_var.set(file_str)

        return file_str

    def get_save_dirpath(self):
        file_str = filedialog.askdirectory(initialdir=self.__base_filepath, mustexist=False)
        if not file_str:
            return

        if self.string_var:
            self.string_var.set(file_str)

        return file_str



"""
Models trained on the 2021 Stanford quantitative double echo
steady state (qDESS) knee dataset.
"""


from copy import deepcopy

import numpy as np

try:
    from keras.layers import BatchNormalization as BN
    from keras.layers import Concatenate, Conv2D, Conv2DTranspose, Dropout, Input, MaxPooling2D
    from keras.models import Model

    _SUPPORTS_KERAS = True
except ImportError:  # pragma: no-cover
    _SUPPORTS_KERAS = False  # pragma: no-cover

from dosma.core.med_volume import MedicalVolume
from dosma.core.orientation import SAGITTAL
from dosma.models.seg_model import KerasSegModel, whiten_volume

__all__ = ["StanfordQDessUNet2D"]


class StanfordQDessUNet2D(KerasSegModel):
    """
    Template for 2D U-Net models trained on the SKM-TEA dataset (previously
    *2021 Stanford qDESS Knee* dataset).

    This model segments patellar cartilage ("pc"), femoral cartilage ("fc"),
    tibial cartilage ("tc"), and the meniscus ("men") from quantitative
    double echo steady state (qDESS) knee scans. The segmentation is computed on
    the root-sum-of-squares (RSS) of the two echoes.

    There are a few weights files that are associated with this model.
    We provide a short description of each below:

        *   ``qDESS_2021_v1-rms-unet2d-pc_fc_tc_men_weights.h5``: This is the baseline
            model trained on the SKM-TEA dataset (v1.0.0).
        *   ``qDESS_2021_v0_0_1-rms-pc_fc_tc_men_weights.h5``: This model is trained on the RSS
            2021 Stanford qDESS knee dataset (v0.0.1).
        *   ``qDESS_2021_v0_0_1-traintest-rms-pc_fc_tc_men_weights.h5``: This model
            is trained on both the train and test set of the 2021 Stanford qDESS knee
            dataset (v0.0.1).

    Examples:

        >>> # Create model based on the volume's shape (SI, AP, 1).
        >>> model = StanfordQDessUNet2D((256, 256, 1), "/path/to/weights")

        >>> # Generate mask from root-sum-of-squares (rss) volume.
        >>> model.generate_mask(rss)

        >>> # Generate mask from dual-echo volume `de_vol` - shape: (SI, AP, LR, 2)
        >>> model.generate_mask(de_vol)
    """

    ALIASES = ("stanford-qdess-2021-unet2d", "skm-tea-unet2d")

    sigmoid_threshold = 0.5

    def __load_keras_model__(self, input_shape):
        if type(input_shape) is not tuple or len(input_shape) != 3 or input_shape[2] != 1:
            raise ValueError("input_size must be a tuple of size (height, width, 1)")

        nfeatures = [2 ** feat * 32 for feat in np.arange(6)]
        depth = len(nfeatures)

        conv_ptr = []

        # input layer
        inputs = Input(input_shape)

        # step down convolutional layers
        pool = inputs
        for depth_cnt in range(depth):

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(pool)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.0)(conv)

            conv_ptr.append(conv)

            # Only maxpool till penultimate depth
            if depth_cnt < depth - 1:

                # If size of input is odd, only do a 3x3 max pool
                xres = conv.shape.as_list()[1]
                if xres % 2 == 0:
                    pooling_size = (2, 2)
                elif xres % 2 == 1:
                    pooling_size = (3, 3)

                pool = MaxPooling2D(pool_size=pooling_size)(conv)

        # step up convolutional layers
        for depth_cnt in range(depth - 2, -1, -1):

            deconv_shape = conv_ptr[depth_cnt].shape.as_list()
            deconv_shape[0] = None

            # If size of input is odd, then do a 3x3 deconv
            if deconv_shape[1] % 2 == 0:
                unpooling_size = (2, 2)
            elif deconv_shape[1] % 2 == 1:
                unpooling_size = (3, 3)

            up = Concatenate(axis=3)(
                [
                    Conv2DTranspose(
                        nfeatures[depth_cnt], (3, 3), padding="same", strides=unpooling_size
                    )(conv),
                    conv_ptr[depth_cnt],
                ]
            )

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(up)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.00)(conv)

        # combine features
        recon = Conv2D(4, (1, 1), padding="same", activation="sigmoid")(conv)

        model = Model(inputs=[inputs], outputs=[recon])

        return model

    def generate_mask(self, volume: MedicalVolume):
        """Segment tissues.

        Args:
            volume (MedicalVolume): The volume to segment. Either 3D or 4D.
                If the volume is 3D, it is assumed to be the root-sum-of-squares (RSS)
                of the two qDESS echoes. If 4D, volume must be of the shape ``(..., 2)``,
                where the last dimension corresponds to echo 1 and 2, respectively.
        """
        ndim = volume.ndim
        if ndim not in (3, 4):
            raise ValueError("`volume` must either be 3D or 4D")

        vol_copy = deepcopy(volume)
        if ndim == 4:
            vol_copy = np.sqrt(np.sum(vol_copy ** 2, axis=-1))

        # reorient to the sagittal plane
        vol_copy.reformat(SAGITTAL, inplace=True)

        vol = vol_copy.volume
        vol = self.__preprocess_volume__(vol)

        # reshape volumes to be (slice, x, y, 1)
        v = np.transpose(vol, (2, 0, 1))
        v = np.expand_dims(v, axis=-1)
        model = self.seg_model

        mask = model.predict(v, batch_size=self.batch_size, verbose=1)
        mask = (mask > self.sigmoid_threshold).astype(np.uint8)

        # reshape mask to be (x, y, slice, classes)
        mask = np.transpose(mask, (1, 2, 0, 3))

        vols = {}
        for i, category in enumerate(["pc", "fc", "tc", "men"]):
            vol_cp = deepcopy(vol_copy)
            vol_cp.volume = mask[..., i]

            # reorient to match with original volume
            vol_cp.reformat(volume.orientation, inplace=True)
            vols[category] = vol_cp

        return vols

    def __preprocess_volume__(self, volume: np.ndarray):
        # TODO: Remove epsilon if difference in performance difference is not large.
        return whiten_volume(volume, eps=1e-8)
